from django.db.models import Sum, Max
from django.http.response import Http404
from django.shortcuts import get_object_or_404
from django.core.exceptions import PermissionDenied
from django.conf import settings
from rest_framework.generics import (
    GenericAPIView, ListAPIView, CreateAPIView, RetrieveAPIView,
    ListCreateAPIView, RetrieveUpdateDestroyAPIView, RetrieveDestroyAPIView,
)
from rest_framework.views import APIView
from rest_framework.parsers import MultiPartParser, FileUploadParser
from rest_framework.response import Response
from rest_framework import status
from rest_framework.exceptions import ValidationError
from arkindex.project.mixins import CorpusACLMixin
from arkindex.project.permissions import IsVerified, IsAuthenticated, IsAdminUser
from arkindex.documents.models import Corpus, Right, Element, ElementType
from arkindex.dataimport.models import \
    DataImport, DataFile, DataImportMode, DataImportFailure, Repository, Event, EventType
from arkindex.dataimport.serializers import (
    DataImportLightSerializer, DataImportSerializer, DataImportFromFilesSerializer,
    DataImportFailureSerializer, DataFileSerializer,
    RepositorySerializer, RepositoryStartImportSerializer,
    ExternalRepositorySerializer, EventSerializer, MLToolSerializer,
)
from arkindex.users.models import OAuthCredentials
from arkindex_common.ml_tool import MLTool
from ponos.models import State
from datetime import datetime
import hashlib
import magic
import os.path


class DataImportsList(CorpusACLMixin, ListAPIView):
    """
    List data imports
    """
    permission_classes = (IsVerified, )
    serializer_class = DataImportLightSerializer

    def get_queryset(self):
        filters = {
            'corpus__in': Corpus.objects.readable(self.request.user)
        }

        if 'corpus' in self.request.query_params:
            filters['corpus'] = self.get_corpus(self.request.query_params['corpus'])

        if 'mode' in self.request.query_params:
            try:
                filters['mode'] = DataImportMode(self.request.query_params['mode'])
            except ValueError:
                raise ValidationError("Mode '{}' does not exist".format(self.request.query_params['mode']))

        if 'id' in self.request.query_params:
            filters['id__startswith'] = self.request.query_params['id']

        qs = DataImport.objects.filter(**filters).prefetch_related('workflow__tasks', 'failures')

        # Order workflow by date of last updated task in workflow
        qs = qs.annotate(last_task=Max('workflow__tasks__updated'), last_run=Max('workflow__tasks__run'))
        return qs.order_by('-last_task')


class DataImportDetails(CorpusACLMixin, RetrieveDestroyAPIView):
    """
    Retrieve and delete a data import
    """

    permission_classes = (IsAuthenticated, )
    serializer_class = DataImportSerializer

    def get_queryset(self):
        return DataImport.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied
        if instance.state == State.Running:
            raise ValidationError("Cannot delete a workflow while it is running")
        super().perform_destroy(instance)


class DataImportRetry(CorpusACLMixin, GenericAPIView):
    """
    Retry a failed import
    """
    permission_classes = (IsVerified, )
    serializer_class = DataImportSerializer

    def get_queryset(self):
        return DataImport.objects.filter(corpus__in=Corpus.objects.writable(self.request.user))

    def post(self, request, *args, **kwargs):
        instance = self.get_object()
        if instance.state in (State.Unscheduled, State.Pending):
            raise ValidationError('This workflow is already pending')
        elif instance.state == State.Running:
            raise ValidationError('This workflow is already running')
        instance.retry()
        serializer = self.get_serializer(instance)
        return Response(serializer.data)


class DataImportFailures(ListAPIView):
    """
    List a single import's errors
    """

    permission_classes = (IsVerified, )
    serializer_class = DataImportFailureSerializer

    def get_queryset(self):
        return DataImportFailure.objects.filter(
            dataimport__corpus__in=Corpus.objects.readable(self.request.user),
            dataimport_id=self.kwargs['pk']
        ).prefetch_related('dataimport__revision__repo', 'element').order_by('path', 'line')


class DataImportFromFiles(CreateAPIView):
    """
    Create, configure and start a workflow from one or more DataFiles
    """

    permission_classes = (IsVerified, )
    serializer_class = DataImportFromFilesSerializer

    def create(self, *args, **kwargs):
        super().create(*args, **kwargs)
        return Response(
            status=status.HTTP_201_CREATED,
            data=DataImportSerializer(self.dataimport, context={'request': self.request}).data,
        )

    def perform_create(self, serializer):
        files = serializer.validated_data['files']
        corpus = files[0].corpus
        mode = serializer.validated_data['mode']
        # Serializer validation codes returns a volume for 'volume_id'
        volume = serializer.validated_data.get('volume_id')
        volume_name = serializer.validated_data.get('volume_name')

        if volume:
            # The files' corpus is already validated as writable
            if volume.corpus != corpus:
                raise ValidationError('Volume and files are in different corpora')
        else:
            if not volume_name:
                if mode == DataImportMode.PDF:
                    suffix = os.path.splitext(files[0].name)[0]
                else:
                    suffix = datetime.today().strftime('%Y-%m-%d %H:%M')
                volume_name = 'Import {} {}'.format(mode.value, suffix)

            volume = corpus.elements.create(
                name=volume_name,
                type=ElementType.Volume,
            )
            volume.events.create(type=EventType.Addition)

        self.dataimport = corpus.imports.create(
            creator=self.request.user,
            mode=mode,
            payload={
                'volume_id': str(volume.id),
                'folder_name': str(volume.id),
            },
        )
        for f in files:
            self.dataimport.files.add(f)
        self.dataimport.start()


class DataImportDemo(CorpusACLMixin, CreateAPIView):
    """
    Create, configure and start an Images workflow from a single DataFile
    """

    permission_classes = (IsAuthenticated, )
    serializer_class = DataImportSerializer

    def create(self, request, pk=None, **kwargs):

        # Get Datafile
        corpora = Corpus.objects.writable(self.request.user)
        datafile = get_object_or_404(DataFile, corpus__in=corpora, id=pk)

        # Get volume, defaults to demo one
        volume_id = self.request.query_params.get('volume')
        filters = {
            'corpus': datafile.corpus,
            'type': ElementType.Volume,
        }
        if volume_id:
            filters['id'] = volume_id
        else:
            filters['name'] = 'Demo'

        volume = get_object_or_404(Element, **filters)

        assert self.has_write_access(volume.corpus), 'Corpus is not writable'

        import_mode = DataImportMode.PDF if datafile.content_type == 'application/pdf' else DataImportMode.Images

        # Start the import
        dataimport = datafile.corpus.imports.create(
            creator=self.request.user,
            mode=import_mode,
            payload={
                'volume_id': str(volume.id),
                'folder_name': 'user-{}-demo'.format(self.request.user.id),
            },
        )
        dataimport.files.add(datafile)
        dataimport.start()

        return Response(data=self.get_serializer(dataimport).data, status=status.HTTP_201_CREATED)


class DataFileList(CorpusACLMixin, ListAPIView):
    """
    List uploaded files
    """
    permission_classes = (IsVerified, )
    serializer_class = DataFileSerializer

    def get_queryset(self):
        return DataFile.objects.filter(corpus=self.get_corpus(self.kwargs['pk']))


class DataFileRetrieve(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Get one file
    """
    permission_classes = (IsVerified, )
    serializer_class = DataFileSerializer

    def get_queryset(self):
        return DataFile.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    def perform_update(self, serializer):
        if not self.has_write_access(serializer.instance.corpus):
            raise PermissionDenied
        return super().perform_update(serializer)

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied
        return super().perform_destroy(instance)


class DataFileUpload(CorpusACLMixin, APIView):
    """
    Upload a new file to a corpus
    """
    permission_classes = (IsAuthenticated, )
    parser_classes = (MultiPartParser, FileUploadParser)

    def post(self, request, pk=None, format=None):
        if 'file' not in request.FILES:
            raise ValidationError({'file': ['No file was sent in the request']})
        try:
            corpus = self.get_corpus(pk, right=Right.Write)
        except Http404:
            raise ValidationError({'corpus': ['Corpus not found']})

        file_obj = request.FILES['file']

        if self.request.user.groups.filter(name="Demo").exists() and \
                corpus.files.exists() and \
                corpus.files.aggregate(Sum('size'))['size__sum'] + file_obj.size > settings.DEMO_FILES_LIMIT:
            raise ValidationError("Corpus total file size limit exceeded")

        md5hash = hashlib.md5()
        for chunk in file_obj.chunks():
            md5hash.update(chunk)
        file_hash = md5hash.hexdigest()

        # Reopen file to reread from beginning
        file_obj.open()
        file_type = magic.from_buffer(file_obj.read(1024), mime=True)

        if DataFile.objects.filter(hash=file_hash, corpus=corpus).exists():
            raise ValidationError('File already exists')

        df = DataFile(
            hash=file_hash,
            corpus=corpus,
            name=file_obj.name,
            size=file_obj.size,
            content_type=file_type,
        )

        file_obj.open()
        df.s3_object.upload_fileobj(file_obj.file)
        df.save()

        return Response(
            data=DataFileSerializer(df).data,
            status=status.HTTP_201_CREATED,
        )


class GitRepositoryImportHook(APIView):
    """
    Handle Git push events
    """

    def post(self, request, pk=None, **kwargs):
        repo = get_object_or_404(Repository, id=pk)
        assert repo.enabled, 'No credentials available'
        repo.provider_class(credentials=repo.credentials).handle_webhook(repo, request)
        return Response(status=status.HTTP_204_NO_CONTENT)


class RepositoryList(ListAPIView):
    permission_classes = (IsVerified, )
    serializer_class = RepositorySerializer

    def get_queryset(self):
        return Repository.objects.filter(
            corpus__in=Corpus.objects.readable(self.request.user),
        )


class AvailableRepositoriesList(ListCreateAPIView):
    permission_classes = (IsVerified, )
    pagination_class = None
    serializer_class = ExternalRepositorySerializer

    def get_queryset(self):
        cred = get_object_or_404(OAuthCredentials, user=self.request.user, id=self.kwargs['pk'])
        return cred.git_provider_class(credentials=cred).list_repos(
            query=self.request.query_params.get('search'),
        )

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        cred = get_object_or_404(OAuthCredentials, user=self.request.user, id=self.kwargs['pk'])

        provider = cred.git_provider_class(credentials=cred)
        repo = provider.create_repo(request=self.request, **serializer.validated_data)
        rev, _ = provider.get_or_create_latest_revision(repo)
        dataimport = rev.start_import()

        return Response(data={'import_id': str(dataimport.id)}, status=status.HTTP_201_CREATED)


class RepositoryRetrieve(CorpusACLMixin, RetrieveDestroyAPIView):
    permission_classes = (IsVerified, )
    serializer_class = RepositorySerializer

    def get_queryset(self):
        return Repository.objects.filter(
            corpus__in=Corpus.objects.readable(self.request.user),
        )

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied
        return super().perform_destroy(instance)


class RepositoryStartImport(RetrieveAPIView):
    permission_classes = (IsAdminUser, )
    serializer_class = RepositoryStartImportSerializer

    def get_queryset(self):
        return Repository.objects.filter(
            corpus__in=Corpus.objects.writable(self.request.user),
        )

    def retrieve(self, request, *args, **kwargs):
        repo = self.get_object()

        if not repo.enabled:
            raise ValidationError("This repository does not have any associated credentials")

        rev, _ = repo.credentials.git_provider_class(
            credentials=repo.credentials,
        ).get_or_create_latest_revision(repo)

        di = rev.start_import()

        return Response(self.get_serializer(di).data)


class ElementHistory(ListAPIView):
    serializer_class = EventSerializer

    def get_queryset(self):
        return Event.objects.filter(
            element_id=self.kwargs['pk'],
            element__corpus__in=Corpus.objects.readable(self.request.user),
        )


class MLToolList(ListAPIView):
    """
    List available machine learning tools
    """
    serializer_class = MLToolSerializer
    pagination_class = None

    def get_queryset(self):
        return sorted(
            MLTool.list(settings.ML_CLASSIFIERS_DIR),
            key=lambda tool: tool.slug,
        )
