import fnmatch
import yaml
from enum import Enum
from django.conf import settings
from django.core.validators import URLValidator
from django.core.exceptions import ValidationError
from django.utils.functional import cached_property
from arkindex_common.ml_tool import MLTool, MLToolType
from arkindex.images.models import ImageServer


class ImportType(Enum):
    Volumes = 'volumes'
    Surfaces = 'surfaces'
    Acts = 'acts'
    Metadata = 'metadata'


class VolumesImportFormat(Enum):
    IIIF = 'iiif'
    TXT = 'txt'


class SurfacesImportFormat(Enum):
    XML = 'xml'


class ActsImportFormat(Enum):
    CSV = 'csv'


class MetadataImportFormat(Enum):
    TEI = 'tei'


class ConfigFile(object):
    """
    A .arkindex.yml configuration file
    """

    FILE_NAME = '.arkindex.yml'
    REQUIRED_ITEMS = {'version', 'branches', 'classifier', 'recognizer'}
    FORMAT_ENUMS = {
        ImportType.Volumes: VolumesImportFormat,
        ImportType.Surfaces: SurfacesImportFormat,
        ImportType.Acts: ActsImportFormat,
        ImportType.Metadata: MetadataImportFormat,
    }

    def __init__(self, data=None):
        if not data:
            return
        if isinstance(data, dict):
            parsed = data
        else:
            parsed = yaml.load(data)
        self.validated_data = self.validate(parsed)
        self.setattrs(self.validated_data)

    @staticmethod
    def from_path(path):
        with open(path, 'rb') as f:
            return ConfigFile(f)

    def validate(self, parsed):
        """
        Validate an Arkindex configuration file. Will raise ValidationErrors for validation errors
        """
        validated_data = {}

        # Required first level items
        for item in self.REQUIRED_ITEMS:
            if item not in parsed:
                raise ValidationError("Missing '{}' setting".format(item))

        # Format version
        if parsed['version'] != 1:
            raise ValidationError("Unsupported format version '{}'".format(parsed['version']))
        validated_data['version'] = parsed['version']

        # Branches list
        if any(not isinstance(branch, str) for branch in parsed['branches']):
            raise ValidationError("Bad 'branches' format: should be a list of branch names")
        validated_data['branches'] = parsed['branches']

        # ML tools
        if parsed['classifier']:
            try:
                MLTool.get(settings.ML_CLASSIFIERS_DIR, MLToolType.Classifier, parsed['classifier'])
            except ValueError:
                raise ValidationError('Classifier "{}" not found'.format(parsed['classifier']))
        validated_data['classifier'] = parsed['classifier']

        if parsed['recognizer']:
            try:
                MLTool.get(settings.ML_CLASSIFIERS_DIR, MLToolType.Recognizer, parsed['recognizer'])
            except ValueError:
                raise ValidationError('Recognizer "{}" not found'.format(parsed['recognizer']))
        validated_data['recognizer'] = parsed['recognizer']

        # At least one import type required
        if not any(it.value in parsed for it in ImportType):
            raise ValidationError("No import types were specified")

        # Required 'paths' for each type
        for it in ImportType:
            if it.value not in parsed:
                continue
            if 'paths' not in parsed[it.value]:
                raise ValidationError("Missing 'paths' parameter in '{}'".format(it.value))
            if not all(isinstance(path, str) for path in parsed[it.value]['paths']):
                raise ValidationError("Bad 'paths' format in '{}': should be a list of patterns".format(it.value))

            validated_data[it.value] = {
                'paths': parsed[it.value]['paths'],
            }

        # Import file formats
        for import_type, format_enum in self.FORMAT_ENUMS.items():
            if import_type.value not in parsed:
                continue
            if 'format' not in parsed[import_type.value]:
                continue
            try:
                format_enum(parsed[import_type.value]['format'])
            except KeyError:
                raise ValidationError("Format setting in '{}' should be one of '{}'".format(
                    import_type.value, "', '".join(fmt.value for fmt in format_enum)))
            validated_data[import_type.value]['format'] = parsed[import_type.value]['format']

        # Manifest import-specific validation
        volumes = parsed.get(ImportType.Volumes.value, {})
        if volumes:
            validated_volumes = {}
            if 'image_servers' in volumes:
                if len(volumes['image_servers'].values()) != len(set(volumes['image_servers'].values())):
                    raise ValueError("Duplicate server URLs in 'volumes.image_servers'")

                validate_url = URLValidator(
                    schemes=['http', 'https'],
                    message="Invalid IIIF server URL in '{}.image_servers'".format(ImportType.Volumes.value),
                )
                for name, url in volumes['image_servers'].items():
                    validate_url(url)

                validated_volumes['image_servers'] = volumes['image_servers']

            if 'lazy_checks' in volumes:
                if not isinstance(volumes['lazy_checks'], bool):
                    raise ValueError("'volumes.lazy_checks' should be true or false")
                validated_volumes['lazy_checks'] = volumes['lazy_checks']

            if 'autoconvert_https' in volumes:
                if not isinstance(volumes['autoconvert_https'], bool):
                    raise ValueError("'volumes.autoconvert_https' should be true or false")
                validated_volumes['autoconvert_https'] = volumes['autoconvert_https']

            validated_data[ImportType.Volumes.value].update(validated_volumes)

        return validated_data

    def setattrs(self, validated_data):
        """
        Set attributes on this class from a valid configration
        """
        self.version = validated_data['version']
        self.branches = validated_data['branches']
        self.imports = list(filter(lambda it: it.value in validated_data, ImportType))
        self.classifier_slug = validated_data['classifier']
        self.recognizer_slug = validated_data['recognizer']

        # Default formats
        self.volumes_format = VolumesImportFormat.IIIF
        self.surfaces_format = SurfacesImportFormat.XML
        self.acts_format = ActsImportFormat.CSV
        self.metadata_format = MetadataImportFormat.TEI

        for import_type, format_enum in self.FORMAT_ENUMS.items():
            if import_type not in self.imports:
                continue
            if 'format' in validated_data[import_type.value]:
                setattr(
                    self,
                    '{}_format'.format(import_type.value),
                    format_enum(validated_data[import_type.value]['format']),
                )
            setattr(
                self,
                '{}_paths'.format(import_type.value),
                validated_data[import_type.value]['paths'],
            )

        if ImportType.Volumes in self.imports:
            vol_conf = validated_data[ImportType.Volumes.value]
            self.volumes_lazy_checks = vol_conf.get('lazy_checks', False)
            self.volumes_autoconvert_https = vol_conf.get('autoconvert_https', False)
            self._volumes_servers = vol_conf.get('image_servers', {})

    def path_match(self, import_type, path):
        """
        Check a given path matches any paths configured for a given import type.
        """
        if import_type not in self.imports:
            return False
        paths = getattr(self, '{}_paths'.format(import_type.value), None)
        return paths and any(fnmatch.fnmatch(path, pattern) for pattern in paths)

    def branch_match(self, branch_name):
        if branch_name.startswith('refs/heads/'):
            branch_name = branch_name[11:]
        return branch_name in self.branches

    @cached_property
    def has_defined_servers(self):
        """
        Returns True if the Volumes import is enabled and custom image servers have been defined.
        """
        return ImportType.Volumes in self.imports and bool(self._volumes_servers)

    @cached_property
    def volumes_image_servers(self):
        if ImportType.Volumes not in self.imports:
            return []

        if not self._volumes_servers:
            return list(ImageServer.objects.all())

        servers = []
        for name, url in self._volumes_servers.items():
            s, _ = ImageServer.objects.get_or_create(url=url, defaults={'display_name': name})
            servers.append(s)
        return servers

    @cached_property
    def classifier(self):
        if not self.classifier_slug:
            return None
        return MLTool.get(settings.ML_CLASSIFIERS_DIR, MLToolType.Classifier, self.classifier_slug)

    @cached_property
    def recognizer(self):
        if not self.recognizer_slug:
            return None
        return MLTool.get(settings.ML_CLASSIFIERS_DIR, MLToolType.Recognizer, self.recognizer_slug)
