from collections import namedtuple
from enum import Enum
from django.core.exceptions import ValidationError
from arkindex.documents.models import Element, ElementType, Page
from arkindex.dataimport.config import ConfigFile, ImportType
from arkindex.dataimport.models import DataImport, DataImportMode
from arkindex.dataimport.filetypes import FileType
from ponos.models import State
import logging
import os
import shutil

logger = logging.getLogger(__name__)


class DiffType(Enum):
    Addition = 'A'
    Modification = 'M'
    Deletion = 'D'
    Rename = 'R'
    Transtype = 'T'
    # The following types exist but should never happen in a simple cloning
    Copy = 'C'
    Unmerged = 'U'
    BrokenPairing = 'B'
    Unknown = 'X'


SimpleDiff = namedtuple('SimpleDiff', 'type, old_path, new_path')


class GitFlow(object):
    '''
    Git repository import workflow in Arkindex
    '''
    def __init__(self, dataimport, working_directory):
        assert isinstance(dataimport, DataImport)
        assert dataimport.mode == DataImportMode.Repository
        assert dataimport.repo is not None
        assert dataimport.sha

        if dataimport.repo.provider_class is None:
            raise ValueError("No repository provider found for {}".format(dataimport.repo.url))

        self.dataimport = dataimport

        assert os.path.isdir(working_directory), \
            'Invalid working directory {}'.format(working_directory)
        self.repo_dir = os.path.join(working_directory, 'repo')
        self.img_dir = os.path.join(working_directory, 'git_images')

    def run(self):
        '''
        Full import workflow
        '''
        self.config = self.parse()
        self.repo = self.clone()
        diffs = self.diff()
        pages = self.dispatch_imports(diffs)
        self.cleanup()
        return pages

    def parse(self):

        logger.info("Fetching configuration data...")
        try:
            config_data = self.dataimport.repo.provider.get_file_content(
                self.dataimport.repo,
                ConfigFile.FILE_NAME,
                ref=self.dataimport.sha,
            )
        except Exception:
            raise Exception('An error occured while fetching configuration data.')

        logger.info("Parsing configuration file...")
        try:
            config = ConfigFile(config_data)
        except (AssertionError, ValidationError) as e:
            raise ValueError('Configuration file validation failed: {}'.format(str(e)))

        logger.info('Configuration file version {}'.format(config.version))
        logger.info('Git branches to trigger imports on: {}'.format(', '.join(config.branches)))
        logger.info('Classifier: {}'.format(config.classifier.name if config.classifier else 'Disabled'))
        logger.info('Recognizer: {}'.format(config.recognizer.name if config.recognizer else 'Disabled'))
        logger.info('Active import types: {}'.format(', '.join(import_type.name for import_type in config.imports)))

        for import_type in config.imports:
            logger.info('Expected format for {} import: {}'.format(
                import_type.name,
                getattr(config, '{}_format'.format(import_type.value)).name,
            ))

            if import_type == ImportType.Volumes:
                if config.has_defined_servers:
                    logger.info('Using image servers {}'.format(
                        ', '.join('{} ({})'.format(srv.display_name, srv.url) for srv in config.volumes_image_servers)
                    ))
                else:
                    logger.info('Using all known image servers')

                logger.info('Image availability checks are {}'.format(
                    'disabled' if config.volumes_lazy_checks else 'enabled'))
                logger.info('HTTPS fallback for image servers is {}'.format(
                    'enabled' if config.volumes_autoconvert_https else 'disabled'))

        rev, _ = self.dataimport.repo.provider.get_or_create_revision(
            self.dataimport.repo,
            self.dataimport.sha,
            save=False,
        )
        assert config.branch_match(rev.ref), \
            'Branch {} does not match configured branches, aborting import'.format(rev.ref)
        rev.save()
        self.dataimport.revision = rev
        self.dataimport.save()

        return config

    def clone(self):
        logger.info('Cloning repo into {}'.format(self.repo_dir))
        assert not os.path.exists(self.repo_dir), \
            'Already existing repo dir {}'.format(self.repo_dir)
        assert self.dataimport.revision is not None

        try:
            repo = self.dataimport.revision.repo.provider.clone_repo(
                self.dataimport.revision.repo,
                self.repo_dir,
                no_checkout=True,
            )
        except Exception as e:
            raise Exception("An error occured while cloning the repository: {}".format(e))

        commit_hash = self.dataimport.revision.hash
        logger.info('Checking out revision {}'.format(commit_hash))
        try:
            repo.head.reference = repo.create_head('commit_{}'.format(commit_hash), commit_hash)
            repo.head.reset(index=True, working_tree=True)
        except Exception as e:
            raise Exception("An error occured while checking out commit {}: {}".format(commit_hash, e))

        return repo

    def diff(self):
        logger.info('Loading Git diff...')

        commit_hash = self.dataimport.revision.hash
        current_commit = self.repo.commit(commit_hash)

        # Iterate over the commit's tree and exclude the current commit to get all parents
        parent_commits = {
            c.hexsha: c
            for c in self.repo.iter_commits(commit_hash)
            if not c.hexsha == commit_hash
        }

        # Look for revisions that match the commit hashes
        parent_revisions = self.dataimport.revision.repo.revisions.filter(
            hash__in=parent_commits.keys()
        ).order_by('-created').prefetch_related('dataimports__workflow__tasks')

        # Get the latest revision where one dataimport was successful
        # Must filter outside of a QuerySet as the DataImport state is a property
        latest_hash = next((
            rev.hash
            for rev in parent_revisions
            if any(d.state == State.Completed for d in rev.dataimports.all())
        ), None)

        if not latest_hash:
            logger.info("No known parent revision found.")
            # No known revision, just return all the repo's files as additions
            # Call git ls-files directly
            return [SimpleDiff(DiffType.Addition, path, path) for path in self.repo.git.ls_files().splitlines()]

        # Pick the Git commit from the latest revision and perform the diff
        diffs = parent_commits[latest_hash].diff(current_commit)

        # Return diff types and paths
        return [SimpleDiff(DiffType(diff.change_type), diff.a_path, diff.b_path) for diff in diffs]

    def dispatch_imports(self, diffs):
        logger.info("Fetching file types...")

        def get_filetype(diff):
            logger.debug("Parsing diff {} from '{}' to '{}'".format(
                diff.type.value, diff.old_path, diff.new_path))
            try:
                return FileType.get(self, diff)
            except Exception as e:
                logger.warning("Error while parsing diff {} from '{}' to '{}': {}".format(
                    diff.type.value, diff.old_path, diff.new_path, str(e)))

        filetypes = FileType.sort(filter(None, map(get_filetype, diffs)))

        changed_objects = []
        for f in filetypes:
            logger.info('Running {}'.format(str(f)))
            try:
                f.handle()
                changed_objects.extend(f.changed_objects)
            except Exception as e:
                logger.warning('Error while running {}: {}'.format(str(f), str(e)))

        def get_page(obj):
            # Pages may be Elements, in which case they must be "cast" to Page
            if isinstance(obj, Element) and obj.type == ElementType.Page:
                obj = obj.page
            if not isinstance(obj, Page):
                return
            return (obj, obj.zone.image.get_thumbnail_url(max_width=None))

        # Changed pages
        return list(filter(None, map(get_page, changed_objects)))

    def cleanup(self):
        logger.info('Cleaning up...')
        shutil.rmtree(self.repo_dir)
