#!/usr/bin/env python3
from django.core.management.base import BaseCommand
from arkindex_common.ml_tool import MLToolType
from arkindex.project.argparse import DataImportArgument
from arkindex.dataimport.models import DataImportMode
from arkindex.dataimport.tasks import (
    download_files,
    check_images,
    extract_pdf_images,
    populate_volume,
    setup_ml_analysis,
)
from arkindex.dataimport.git import GitFlow
from django.conf import settings
import tempfile
import json
import os
import logging
import shutil

logging.basicConfig(
    level=logging.INFO,
    format='[%(levelname)s] %(message)s',
)
logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Import files from a data import'

    def add_arguments(self, parser):
        parser.add_argument(
            'data_import',
            type=DataImportArgument(),
            help='ID of the DataImport to run'
        )

    def handle(self, *args, data_import=None, **options):
        # Use default ML tools here
        ml_tools = (
            (MLToolType.Classifier, settings.ML_DEFAULT_CLASSIFIER),
            (MLToolType.Recognizer, settings.ML_DEFAULT_RECOGNIZER),
        )

        # Use shared directory when running in docker
        # Fallback to a temp directory while developing
        task_dir = os.environ.get('PONOS_DATA', tempfile.mkdtemp(suffix='-ponos'))
        assert os.path.isdir(task_dir), 'Invalid task dir {}'.format(task_dir)
        logger.info('Using task dir: {}'.format(task_dir))

        # Use temp folder for anything that does not need to be shared
        temp_dir = tempfile.mkdtemp(suffix='-ponostmp')
        logger.debug('Using temp dir: {}'.format(temp_dir))

        pages, files = None, None
        if data_import.files.exists():
            files = download_files(data_import, temp_dir)

        if data_import.mode == DataImportMode.Images:
            # Validate images from data import
            files = check_images(files)

        elif data_import.mode == DataImportMode.PDF:
            assert len(files) == 1, 'Only one file in PDF mode'
            pdf_file, pdf_path = files[0]
            assert pdf_file.content_type == 'application/pdf', 'File is not a PDF'

            # Extract images from the PDF into the task working dir and get their paths
            images = extract_pdf_images(pdf_file, pdf_path, task_dir)

            # Add those images to the volume
            # all linked to the original pdf file
            files = [
                (pdf_file, img_path)
                for img_path in images
            ]

        elif data_import.mode == DataImportMode.Repository:
            git = GitFlow(data_import, task_dir)
            pages = git.run()

            # use ml_tools from Repo
            ml_tools = []
            if git.config.classifier:
                ml_tools.append((MLToolType.Classifier, git.config.classifier.slug))
            if git.config.recognizer:
                ml_tools.append((MLToolType.Recognizer, git.config.recognizer.slug))

        else:
            # Should never happen
            raise NotImplementedError

        # Load all image files into the volume
        if pages is None and files is not None:

            # Load or create volume from dataimport
            volume = data_import.get_volume()
            logger.info('Using volume: {}'.format(volume))

            pages = populate_volume(volume, files)

        shutil.rmtree(temp_dir)

        # Setup the analysis process through a json config
        ml_analysis = os.path.join(task_dir, 'ml_analysis.json')
        if os.path.exists(ml_analysis) or not ml_tools:
            return
        with open(ml_analysis, 'w') as f:
            json.dump(setup_ml_analysis(pages, ml_tools), f, indent=4)
