from django.db import models
from django.contrib.postgres.fields import JSONField
from django.template.defaultfilters import slugify
from django.conf import settings
from django.utils.functional import cached_property
from rest_framework.exceptions import ValidationError
from enumfields import EnumField, Enum
from arkindex.project.aws import s3
from arkindex.project.models import IndexableModel
from arkindex.documents.models import Element, ElementType
from arkindex.dataimport.providers import git_providers, get_provider
from ponos.models import Workflow, State
from botocore.exceptions import ClientError
from io import BytesIO
import logging
import os
import yaml
import uuid

logger = logging.getLogger(__name__)


class DataImportMode(Enum):
    Images = 'images'
    PDF = 'pdf'
    Repository = 'repository'


class DataImport(IndexableModel):
    """
    A single import workflow
    """

    creator = models.ForeignKey('users.User', on_delete=models.CASCADE, related_name='imports')
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='imports')
    mode = EnumField(DataImportMode, max_length=30)
    files = models.ManyToManyField('dataimport.DataFile', related_name='imports')
    revision = models.ForeignKey(
        'dataimport.Revision', related_name='dataimports', on_delete=models.CASCADE, blank=True, null=True)
    payload = JSONField(null=True, blank=True)
    workflow = models.OneToOneField('ponos.Workflow', on_delete=models.SET_NULL, null=True, blank=True)

    class Meta:
        ordering = ['corpus', '-created']

    @property
    def state(self):
        if not self.workflow:
            return State.Unscheduled
        # This allows annotating a DataImport queryset with "last_run" and preventing duplicate SQL queries
        if hasattr(self, 'last_run'):
            return self.workflow.get_state(self.last_run)
        else:
            return self.workflow.state

    def build_workflow(self):
        '''
        Create a ponos workflow with a recipe according to configuration
        '''
        assert self.workflow is None, 'A workflow is already setup'
        if self.mode == DataImportMode.Repository and self.revision is not None and not self.revision.repo.enabled:
            raise ValidationError('Git repository does not have any valid credentials')

        # Import data in Arkindex, then do the ML analysis
        tasks = {
            'import': {
                'image': settings.ARKINDEX_APP_IMAGE,
                'command': 'manage.py import {}'.format(self.id),
            },
            'ml': {
                'image': settings.ARKINDEX_ML_IMAGE,
                'command': 'arkindex-ml-analysis /data/ml_analysis.json',
                'parent': 'import',
            },
        }

        # Build the workflow in db
        recipe = settings.PONOS_RECIPE.copy()
        recipe['tasks'] = tasks
        return Workflow.objects.create(recipe=yaml.dump(recipe))

    def start(self):
        self.workflow = self.build_workflow()
        self.save()

    def retry(self):
        if self.mode == DataImportMode.Repository and self.revision is not None and not self.revision.repo.enabled:
            raise ValidationError('Git repository does not have any valid credentials')
        if self.workflow:
            self.workflow.retry()
        else:
            self.start()

    def get_volume(self):
        '''
        Get or create the associated volume
        '''
        volume_id = self.payload.get('volume_id')
        volume_name = self.payload.get('volume_name')
        if volume_id:
            # Get existing volume
            try:
                return Element.objects.get(
                    type=ElementType.Volume,
                    corpus=self.corpus,
                    pk=volume_id,
                )
            except Element.DoesNotExist:
                raise Exception('Missing volume on data import')

        elif volume_name:
            # Create a new volume
            return Element.objects.create(
                type=ElementType.Volume,
                corpus=self.corpus,
                name=volume_name,
            )

        else:
            raise Exception('No volume available')

    @cached_property
    def folder_name(self):
        if 'folder_name' not in self.payload:
            return
        return slugify(self.payload['folder_name'])

    @cached_property
    def iiif_path(self):
        if not self.folder_name:
            return
        return os.path.join(settings.LOCAL_IMAGESERVER_ROOT, self.folder_name)

    @cached_property
    def repo(self):
        if self.revision:
            return self.revision.repo
        elif 'repo_id' in self.payload:
            return Repository.objects.get(id=self.payload['repo_id'])

    @cached_property
    def sha(self):
        return self.revision.hash if self.revision else self.payload.get('sha')


class DataImportFailure(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    dataimport = models.ForeignKey(
        'dataimport.DataImport', on_delete=models.CASCADE, related_name='failures')
    element = models.ForeignKey(
        'documents.Element', on_delete=models.CASCADE, blank=True, null=True, related_name='failures')
    path = models.CharField(max_length=250)
    line = models.PositiveIntegerField(blank=True, null=True)
    message = models.TextField()
    context = models.TextField(blank=True, null=True)

    @property
    def view_url(self):
        if not self.line or not self.dataimport.revision:
            return None
        return '{}/blob/{}/{}#L{}'.format(
            self.dataimport.revision.repo.url.rstrip('/'),
            self.dataimport.revision.hash,
            self.path,
            self.line,
        )


class DataFile(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    name = models.CharField(max_length=100)
    size = models.PositiveIntegerField()
    hash = models.CharField(max_length=32)
    content_type = models.CharField(max_length=50)
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='files')
    image = models.OneToOneField('images.Image', on_delete=models.SET_NULL, blank=True, null=True)

    class Meta:
        unique_together = (('corpus', 'hash'), )
        ordering = ['corpus', 'name']

    @property
    def staging_path(self):
        return str(self.id)

    @cached_property
    def s3_object(self):
        return s3.Object(settings.AWS_STAGING_BUCKET, self.staging_path)

    def exists(self):
        try:
            self.s3_object.load()
            return True
        except ClientError as e:
            if e.response['Error']['Code'] != '404':
                raise
            return False

    def download(self):
        b = BytesIO()
        logger.debug('Downloading file {} from S3'.format(self.staging_path))
        self.s3_object.download_fileobj(b)
        return b

    def download_to(self, path):
        logger.debug('Downloading file {} from S3'.format(self.staging_path))
        self.s3_object.download_file(path)


class Repository(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    url = models.URLField(unique=True)
    hook_token = models.CharField(max_length=250, unique=True)
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='repos')
    credentials = models.ForeignKey(
        'users.OAuthCredentials', on_delete=models.SET_NULL, related_name='repos', blank=True, null=True)
    provider_name = models.CharField(
        max_length=50,
        choices=[(p.__name__, p.display_name) for p in git_providers],
    )

    class Meta:
        verbose_name_plural = 'repositories'

    @property
    def provider_class(self):
        return get_provider(self.provider_name)

    @property
    def provider(self):
        return self.provider_class(credentials=self.credentials)

    @property
    def enabled(self):
        return self.credentials is not None and self.provider_class is not None


class Revision(IndexableModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    repo = models.ForeignKey('dataimport.Repository', on_delete=models.CASCADE, related_name='revisions')
    hash = models.CharField(max_length=50)
    ref = models.CharField(max_length=50)
    message = models.TextField()
    author = models.CharField(max_length=50)
    elements = models.ManyToManyField('documents.Element', through='dataimport.Event', related_name='revisions')

    class Meta:
        unique_together = (('repo', 'hash'), )

    @property
    def commit_url(self):
        return '{}/commit/{}'.format(self.repo.url.rstrip('/'), self.hash)

    def start_import(self):
        dataimport = self.dataimports.create(
            creator=self.repo.credentials.user,
            corpus=self.repo.corpus,
            mode=DataImportMode.Repository,
            payload={
                'repo_id': str(self.repo.id),
                'sha': self.hash,
            }
        )
        dataimport.start()
        return dataimport

    def __str__(self):
        return '{} "{}" by {}'.format(self.hash[:8], self.message.splitlines()[0], self.author)


class EventType(Enum):
    Addition = 'A'
    Edit = 'M'
    Deletion = 'D'


class Event(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    element = models.ForeignKey('documents.Element', on_delete=models.CASCADE, related_name='events')
    revision = models.ForeignKey(
        'dataimport.Revision', on_delete=models.CASCADE, related_name='events', blank=True, null=True)
    type = EnumField(EventType, max_length=10)

    class Meta:
        unique_together = (('element', 'revision'), )
        ordering = ['element_id', '-created']
