from abc import ABC, abstractmethod
from django.urls import reverse
from rest_framework.exceptions import NotAuthenticated, AuthenticationFailed, APIException, ValidationError
from gitlab import Gitlab, GitlabGetError, GitlabCreateError
from arkindex.documents.models import Corpus
import urllib.parse
import base64
import uuid
import git


class GitProvider(ABC):

    display_name = None

    def __init__(self, credentials=None, url=None):
        if credentials:
            from arkindex.users.models import OAuthCredentials
            assert isinstance(credentials, OAuthCredentials)
            self.url = credentials.provider_url
        if url:
            self.url = url
        self.credentials = credentials

    @abstractmethod
    def list_repos(self, query=None):
        """
        List all repositories or filter with a search query.
        """

    @abstractmethod
    def create_repo(self, **kwargs):
        """
        Create a Repository instance from an external repository
        """

    @abstractmethod
    def clone_repo(self, repo, dest_dir, **kwargs):
        """
        Get a git.Repo instance for a repository cloned in a given destination directory.
        """

    @abstractmethod
    def download_archive(self, revision, path):
        """
        Download an archive for a given Revision instance.
        """

    def get_or_create_revision(self, repo, sha, save=True):
        from arkindex.dataimport.models import Revision
        try:
            return self.get_revision(repo, sha), False
        except Revision.DoesNotExist:
            return self.create_revision(repo, sha, save=save), True

    def get_revision(self, repo, sha):
        return repo.revisions.get(hash=sha)

    @abstractmethod
    def create_revision(self, repo, sha, save=True):
        """
        Create a Revision instance for a given commit hash of a given repository.
        """

    @abstractmethod
    def get_or_create_latest_revision(self, repo):
        """
        Get or create a Revision instance for the last revision on the main branch of a given repository.
        """

    @abstractmethod
    def get_file_content(self, repo, path, ref="master"):
        """
        Get the contents of a given file on a given repository.
        """

    @abstractmethod
    def handle_webhook(self, repo, request):
        """
        Handle a webhook event on a given repository.
        """


class GitLabProvider(GitProvider):

    display_name = "GitLab"
    url = 'https://gitlab.com'

    def _try_get_project(self, gl, id):
        try:
            return gl.projects.get(id)
        except GitlabGetError as e:
            raise APIException("Error while fetching GitLab project: {}".format(str(e)))

    def list_repos(self, query=None):
        if not self.credentials:
            raise NotAuthenticated()
        gl = Gitlab(self.url, oauth_token=self.credentials.token)
        return gl.projects.list(membership=True, search=query)

    def create_repo(self, id=None, corpus=None, request=None, **kwargs):
        assert isinstance(corpus, Corpus)
        if not self.credentials and request:
            raise NotAuthenticated()
        gl = Gitlab(self.url, oauth_token=self.credentials.token)
        project = self._try_get_project(gl, int(id))

        from arkindex.dataimport.models import Repository
        if Repository.objects.filter(url=project.web_url).exists():
            raise ValidationError("A repository with this URL already exists")

        repo = self.credentials.repos.create(
            corpus=corpus,
            url=project.web_url,
            hook_token=base64.b64encode(uuid.uuid4().bytes).decode('utf-8'),
            provider_name=self.__class__.__name__,
        )

        try:
            project.hooks.create({
                'url': request.build_absolute_uri(
                    reverse('api:import-hook', kwargs={'pk': repo.id}),
                ),
                'push_events': True,
                'token': repo.hook_token,
            })
        except GitlabCreateError as e:
            raise APIException("Error while creating GitLab hook: {}".format(str(e)))

        return repo

    def clone_repo(self, repo, dest_dir, **kwargs):
        parsed = list(urllib.parse.urlsplit(repo.url))

        # Clone over HTTPS using an OAuth token is an undocumented feature supported since GitLab 8.12
        # https://gitlab.com/gitlab-org/gitlab-ce/merge_requests/10677
        parsed[1] = 'oauth2:{}@{}'.format(repo.credentials.token, parsed[1])

        return git.Repo.clone_from(urllib.parse.urlunsplit(parsed), dest_dir, **kwargs)

    def download_archive(self, revision, path):
        gl = Gitlab(self.url, oauth_token=revision.repo.credentials.token)
        project = self._try_get_project(gl, urllib.parse.urlsplit(revision.repo.url).path.strip('/'))

        with open(path, 'wb') as f:
            project.repository_archive(sha=revision.hash, streamed=True, action=f.write)

    def create_revision(self, repo, sha, save=True):
        from arkindex.dataimport.models import Revision

        gl = Gitlab(self.url, oauth_token=repo.credentials.token)
        project = self._try_get_project(gl, urllib.parse.urlsplit(repo.url).path.strip('/'))
        commit = project.commits.get(sha)
        branch = next(ref['name'] for ref in commit.refs() if ref['type'] == 'branch')

        rev = Revision(
            repo=repo,
            hash=sha,
            ref=branch,
            message=commit.message,
            author=commit.author_name,
        )
        if save:
            rev.save()
        return rev

    def get_or_create_latest_revision(self, repo):
        gl = Gitlab(self.url, oauth_token=repo.credentials.token)
        project = self._try_get_project(gl, urllib.parse.urlsplit(repo.url).path.strip('/'))

        latest_commit = project.commits.list()[0]
        return repo.revisions.get_or_create(
            hash=latest_commit.id,
            defaults={
                'ref': latest_commit.refs()[0]['name'],
                'message': latest_commit.message,
                'author': latest_commit.author_name,
            },
        )

    def get_file_content(self, repo, path, ref="master"):
        gl = Gitlab(self.url, oauth_token=repo.credentials.token)
        project = self._try_get_project(gl, urllib.parse.urlsplit(repo.url).path.strip('/'))

        return project.files.get(file_path=path, ref=ref).decode()

    def handle_webhook(self, repo, request):
        from arkindex.dataimport.models import DataImport, DataImportMode

        if 'HTTP_X_GITLAB_EVENT' not in request.META:
            raise ValidationError("Missing GitLab event type")
        if request.META['HTTP_X_GITLAB_EVENT'] != 'Push Hook':
            raise ValidationError("Unsupported GitLab event type")

        if 'HTTP_X_GITLAB_TOKEN' not in request.META:
            raise NotAuthenticated("Missing GitLab secret token")
        if request.META['HTTP_X_GITLAB_TOKEN'] != repo.hook_token:
            raise AuthenticationFailed("Invalid GitLab secret token")

        if not isinstance(request.data, dict) or 'object_kind' not in request.data:
            raise ValidationError('Bad payload format')

        if not request.data['object_kind'] == 'push':
            raise ValidationError("Unsupported GitLab event type")

        sha = request.data.get('checkout_sha')
        if not sha:
            raise ValidationError('Missing checkout SHA')

        # Already took care of this event
        if repo.revisions.filter(hash=sha).exists():
            return

        di = DataImport.objects.create(
            corpus=repo.corpus,
            creator=repo.credentials.user,
            mode=DataImportMode.Repository,
            payload={
                'repo_id': str(repo.id),
                'sha': sha,
            },
        )
        di.start()


git_providers = [
    GitLabProvider,
]
oauth_to_git = {
    "gitlab": GitLabProvider,
}


def get_provider(name):
    return next(filter(lambda p: p.__name__ == name, git_providers), None)


def from_oauth(name):
    return oauth_to_git.get(name)
