from rest_framework import serializers
from rest_framework.utils import model_meta
from arkindex.project.serializer_fields import EnumField
from arkindex.dataimport.models import (
    DataImport, DataImportMode, DataImportFailure, DataFile,
    Repository, Revision, Event, EventType
)
from arkindex.documents.models import Corpus, Element, ElementType
from arkindex.documents.serializers.light import ElementLightSerializer
from arkindex_common.ml_tool import MLToolType
from ponos.models import State
import gitlab.v4.objects


class DataImportLightSerializer(serializers.ModelSerializer):
    """
    Serialize a data importing workflow
    """

    state = EnumField(State)
    mode = EnumField(DataImportMode)
    creator = serializers.HiddenField(default=serializers.CurrentUserDefault())
    has_failures = serializers.SerializerMethodField()

    def get_has_failures(self, obj):
        return obj.failures.exists()

    class Meta:
        model = DataImport
        fields = (
            'id',
            'state',
            'mode',
            'corpus',
            'creator',
            'has_failures',
        )
        read_only_fields = ('id', 'state')


class FilesPayloadSerializer(serializers.Serializer):
    """
    Serialize an image importing payload
    """

    folder_name = serializers.CharField()
    volume_id = serializers.CharField()


class RevisionSerializer(serializers.ModelSerializer):
    """
    Serialize a repository revision
    """

    date = serializers.DateTimeField(source='created')

    class Meta:
        model = Revision
        fields = (
            'id',
            'date',
            'hash',
            'ref',
            'message',
            'author',
            'commit_url',
            'repo_id',
        )


class DataImportSerializer(DataImportLightSerializer):
    """
    Serialize a data importing workflow with its payload
    """

    # Redefine state as writable
    state = EnumField(State)
    payload = serializers.JSONField()
    revision = RevisionSerializer(read_only=True)
    workflow = serializers.HyperlinkedRelatedField(read_only=True, view_name='ponos:workflow-details')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance is None:
            return
        payload_serializers = {
            DataImportMode.Images: FilesPayloadSerializer,
            DataImportMode.PDF: FilesPayloadSerializer,
        }
        self.fields['payload'] = payload_serializers.get(self.instance.mode, serializers.JSONField)()

    def update(self, instance, validated_data):
        # ModelSerializer disallows updating with nested serializers.
        # We dynamically inject a payload serializer, but it is useless for the database
        # since the payload field is a JSONField with no particular checks.
        # We don't have to care about a nested serializer and we can treat it just as a dict.
        # This code is copy-pasted from DRF's ModelSerializer.
        info = model_meta.get_field_info(instance)
        for attr, value in validated_data.items():
            if attr in info.relations and info.relations[attr].to_many:
                field = getattr(instance, attr)
                field.set(value)
            else:
                setattr(instance, attr, value)
        instance.save()
        return instance

    class Meta(DataImportLightSerializer.Meta):
        fields = DataImportLightSerializer.Meta.fields + (
            'payload',
            'files',
            'revision',
            'workflow',
        )


class DataImportFromFilesSerializer(serializers.Serializer):

    mode = EnumField(DataImportMode, default=DataImportMode.Images)
    files = serializers.PrimaryKeyRelatedField(queryset=DataFile.objects.all(), many=True)
    volume_id = serializers.UUIDField(required=False, allow_null=True)
    volume_name = serializers.CharField(max_length=250, required=False, allow_null=True)

    def validate_mode(self, mode):
        if mode not in (DataImportMode.Images, DataImportMode.PDF):
            raise serializers.ValidationError("This mode is not allowed when importing from files")
        return mode

    def validate_files(self, files):
        if not files:
            raise serializers.ValidationError('At least one file is required to start an import')
        corpora = set(f.corpus for f in files)
        if len(corpora) > 1:
            raise serializers.ValidationError('Imports can only run on files from a single corpus')
        corpus = corpora.pop()
        if corpus not in Corpus.objects.writable(self.context['request'].user):
            raise serializers.ValidationError('Cannot write in corpus')
        return files

    def validate_volume_id(self, volume_id):
        if not volume_id:
            return None
        try:
            volume = Element.objects.get(
                id=volume_id,
                type=ElementType.Volume,
            )
            return volume
        except Element.DoesNotExist:
            raise serializers.ValidationError('Volume does not exist')

    def validate(self, data):
        if data['mode'] == DataImportMode.PDF:
            if len(data['files']) > 1:
                raise serializers.ValidationError('Cannot import more than one PDF at once')
            if data['files'][0].content_type != 'application/pdf':
                raise serializers.ValidationError('PDF imports can only import PDF files')

        elif data['mode'] == DataImportMode.Images:
            if not all(f.content_type.startswith('image/') for f in data['files']):
                raise serializers.ValidationError('Image imports can only import image files')

        else:
            raise NotImplementedError

        return data


class DataFileSerializer(serializers.ModelSerializer):
    """
    Serialize a single uploaded file
    """

    class Meta:
        model = DataFile
        fields = (
            'id',
            'name',
            'content_type',
            'size',
        )
        read_only_fields = ('id', 'size', 'content_type', )


class DataImportFailureSerializer(serializers.ModelSerializer):
    """
    Serialize a data import error log
    """

    element = ElementLightSerializer()

    class Meta:
        model = DataImportFailure
        fields = (
            'id',
            'element',
            'path',
            'line',
            'message',
            'context',
            'view_url',
        )


class RepositorySerializer(serializers.ModelSerializer):
    """
    Serialize a repository
    """

    class Meta:
        model = Repository
        fields = (
            'id',
            'url',
            'enabled',
            'corpus',
        )
        extra_kwargs = {
            'id': {'read_only': True},
            'url': {'read_only': True},
            'enabled': {'read_only': True},
        }


class RepositoryStartImportSerializer(serializers.ModelSerializer):
    """
    A serializer used by the RepositoryStartImport endpoint to return a DataImport ID.
    This serializer is required to get the OpenAPI schema generation to work.
    """

    import_id = serializers.UUIDField(source='id')

    class Meta:
        model = DataImport
        fields = ('import_id',)
        read_only_fields = ('import_id',)


class ExternalRepositorySerializer(serializers.Serializer):
    """
    Serialize a Git repository from an external API
    """

    # Useless fields, added to prevent 500 errors when opening
    # an API endpoint in the browser
    id = serializers.IntegerField(min_value=0)
    corpus = serializers.UUIDField()

    def to_representation(self, obj):
        if isinstance(obj, gitlab.v4.objects.Project):
            return {
                "id": obj.id,
                "name": obj.name_with_namespace,
                "url": obj.web_url,
            }
        else:
            raise NotImplementedError

    def to_internal_value(self, data):
        """
        Deserializing only requires a 'id' attribute
        """
        if not data.get('id'):
            raise serializers.ValidationError({
                'id': 'This field is required.'
            })
        if not data.get('corpus'):
            raise serializers.ValidationError({
                'corpus': 'This field is required.'
            })

        return {
            'id': data['id'],
            'corpus': Corpus.objects.writable(self.context['request'].user)
                                    .get(id=data['corpus'])
        }


class EventSerializer(serializers.ModelSerializer):
    """
    Serialize a diff event for an element on a revision
    """

    type = EnumField(EventType)
    revision = RevisionSerializer()
    date = serializers.DateTimeField(source='created')

    class Meta:
        model = Event
        fields = (
            'id',
            'type',
            'date',
            'revision',
        )


class MLToolSerializer(serializers.Serializer):
    """
    Serialize a machine learning tool for display
    """
    name = serializers.CharField()
    slug = serializers.SlugField()
    type = EnumField(MLToolType)
    version = serializers.CharField()
