from arkindex.documents.models import Element, ElementType
from arkindex.documents.importer import import_page
from arkindex.dataimport.tasks.image import build_iiif_image
from arkindex.dataimport.models import EventType, DataImport
from botocore.exceptions import ClientError
import logging
import os.path

logger = logging.getLogger(__name__)


def download_files(dataimport, dest_dir):
    assert isinstance(dataimport, DataImport)
    assert os.access(dest_dir, os.W_OK | os.X_OK), 'Destination directory is read-only'

    datafiles = dataimport.files.all()
    valid_files = []
    filecount = len(datafiles)

    if filecount < 1:
        logger.info('No files to download - skipping')
        return []

    for i, datafile in enumerate(datafiles):
        logger.info('Downloading file {} of {}'.format(i + 1, filecount))
        path = os.path.join(dest_dir, datafile.staging_path)

        try:
            datafile.download_to(path)
        except (IOError, ClientError) as e:
            logger.warning('Failed downloading file {} ({}): {}'.format(datafile.name, str(datafile.id), str(e)))
            continue

        valid_files.append((datafile, path))

    assert len(valid_files) > 0, 'No files were successfully downloaded'
    return valid_files


def populate_volume(volume, files):
    '''
    Import files into the volume, and post on IIIF server
    TODO: this could use an API endpoint to ingest a new page
    '''
    logger.info('Pre import checks...')
    assert len(files) > 0, 'No files to import'
    assert isinstance(volume, Element)
    assert volume.type == ElementType.Volume

    pages, count = [], len(files)
    for i, (data_file, staging_path) in enumerate(files):
        logger.info('Adding page {} of {}'.format(i + 1, count))

        # Build local IIIF image
        img = build_iiif_image(volume, staging_path, data_file, suffix=str(i))

        # Build page with image
        page = import_page(volume, img, volume.name)
        page.events.create(type=EventType.Addition)
        pages.append((page, img.s3_url))

    assert len(pages) > 0, 'No imported pages'
    logger.info("Imported {} pages into {}".format(len(pages), volume.name))
    volume.generate_thumbnail()

    return pages


def setup_ml_analysis(pages, tools):
    '''
    Build a JSON serializable configuration for ML analysis
    Using the structure from populate_volume
    '''
    return {
        'pages': [
            {
                'id': str(page.id),
                'path': path,
            }
            for page, path in pages
        ],
        'tools': [
            {
                'tool': tool.value,
                'slug': slug,
            }
            for tool, slug in tools
        ]
    }
