from PIL import Image as PillowImage
from arkindex.dataimport.models import DataFile
from arkindex.images.models import ImageServer, Image, ImageStatus
from arkindex.documents.models import Element, ElementType
from urllib.parse import quote
import logging

logger = logging.getLogger(__name__)


def check_images(files):
    assert len(files), 'No files to check'

    valid_files = []
    filecount = len(files)

    for i, (datafile, path) in enumerate(files):
        logger.info("Checking image {} of {}".format(i + 1, filecount))

        try:
            img = PillowImage.open(path)
            assert max(img.size) >= 500, "Image {} is too small".format(datafile.name)
        except IOError:
            logger.warn("File {} is not a valid image".format(datafile.name))
            continue
        except AssertionError as e:
            logger.warn(str(e))
            continue

        valid_files.append((datafile, path))

    assert len(valid_files) > 0, "No valid images in selected files"
    return valid_files


def build_iiif_image(volume, path, data_file, suffix=None):
    '''
    Import a staging image into the local IIIF server
    '''
    assert isinstance(volume, Element)
    assert volume.type == ElementType.Volume
    assert isinstance(path, str)
    assert isinstance(data_file, DataFile)

    pillow_img = PillowImage.open(path)
    width, height = pillow_img.size

    # Non-JPEG image formats that should not be converted
    # Will default to .jpg if the image format is not in there
    # Formats are Pillow images formats,
    # see https://pillow.readthedocs.io/en/5.1.x/handbook/image-file-formats.html
    known_exts = {
        "JPEG2000": ".jp2",
        "PNG": ".png",
        "TIFF": ".tif",
    }
    img_format = pillow_img.format
    if img_format not in known_exts:
        img_format = 'JPEG'
    ext = known_exts.get(img_format, '.jpg')

    # Build image path
    filename = str(data_file.id)
    if suffix is not None:
        filename += '-{}'.format(suffix)
    filename += ext
    iiif_path = quote('{}/{}'.format(volume.id, filename), safe='')

    # Get Image instance
    try:
        img = ImageServer.objects.local.images.get(path=iiif_path)
    except Image.DoesNotExist:
        img = Image(
            server=ImageServer.objects.local,
            path=iiif_path,
            width=width,
            height=height,
            datafile=data_file,
        )

    if img.exists():
        logger.warning('Image already exists on the IIIF server')
    else:
        # Save to S3 using optional image type conversion
        img.pillow_save(pillow_img, format=img_format)
        img.status = ImageStatus.Checked
        img.save()

    return img
