from unittest.mock import patch, call
from rest_framework import status
from django.urls import reverse
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import override_settings
from arkindex.documents.models import Corpus
from arkindex.dataimport.models import DataFile
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import User
import uuid


@override_settings(DEMO_GROUP_ID=2, AWS_STAGING_BUCKET='staging')
class TestFiles(FixtureAPITestCase):
    """
    Test file management
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.demo_user = User.objects.get(email='demo@demo.fr')
        cls.demo_user.corpus_right.create(corpus=cls.corpus, can_write=True)

    def setUp(self):
        super().setUp()
        self.client.force_login(self.user)

    def test_file_list(self):
        df = DataFile.objects.create(
            name='test.txt', size=42, hash='aaaa', content_type='text/plain', corpus=self.corpus)

        response = self.client.get(reverse('api:file-list', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertIn('results', data)
        self.assertEqual(len(data['results']), 1)

        file = data['results'][0]
        self.assertIn('id', file)
        self.assertIn('name', file)
        self.assertIn('size', file)
        self.assertEqual(file['id'], str(df.id))
        self.assertEqual(file['name'], df.name)
        self.assertEqual(file['size'], df.size)

    def test_file_list_requires_login(self):
        self.client.logout()
        response = self.client.get(reverse('api:file-list', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_file_upload_ro(self):
        '''
        An upload on a read only corpus should fail
        '''
        public = Corpus.objects.create(name='Public', public=True)
        f = SimpleUploadedFile('test.txt', b'This is a text file')

        response = self.client.post(reverse('api:file-upload', kwargs={'pk': public.id}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch('arkindex.dataimport.models.s3')
    def test_file_upload(self, s3_mock):
        """
        Assert a file upload creates a database instance and saves the file
        """
        f = SimpleUploadedFile('test.txt', b'This is a text file')

        response = self.client.post(reverse('api:file-upload', kwargs={'pk': self.corpus.id}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn('id', data)

        df = DataFile.objects.get(id=data['id'])
        self.assertEqual(df.name, 'test.txt')
        self.assertEqual(df.size, 19)
        self.assertEqual(df.content_type, 'text/plain')
        self.assertEqual(df.corpus, self.corpus)
        self.assertEqual(s3_mock.Object.call_count, 1)
        self.assertEqual(s3_mock.Object.call_args, call('staging', str(df.id)))
        self.assertEqual(s3_mock.Object().upload_fileobj.call_count, 1)

    def test_file_upload_requires_login(self):
        """
        Assert a file upload requires a login
        """
        self.client.logout()
        f = SimpleUploadedFile('test.txt', b'This is a text file')
        response = self.client.post(reverse('api:file-upload', kwargs={'pk': self.corpus.id}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_file_upload_unknown_corpus(self):
        """
        Assert file upload fails on an unknown corpus with HTTP 400
        """
        f = SimpleUploadedFile('test.txt', b'This is a text file')
        response = self.client.post(reverse('api:file-upload', kwargs={'pk': str(uuid.uuid4())}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    @patch('arkindex.dataimport.models.s3')
    def test_file_upload_wrong_content_type(self, s3_mock):
        """
        Assert file upload does not trust the client defined content type
        """
        f = SimpleUploadedFile('test.mp3', b'This actually is a text file', content_type='audio/mpeg')

        response = self.client.post(reverse('api:file-upload', kwargs={'pk': self.corpus.id}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn('id', data)

        df = DataFile.objects.get(id=data['id'])
        self.assertEqual(df.name, 'test.mp3')
        self.assertEqual(df.content_type, 'text/plain')

        self.assertEqual(s3_mock.Object.call_count, 1)
        self.assertEqual(s3_mock.Object().upload_fileobj.call_count, 1)

    @patch('arkindex.dataimport.models.s3')
    def test_file_upload_unique(self, s3_mock):
        """
        Assert uploading the same file twice fails
        """
        f = SimpleUploadedFile('test.txt', b'This is a text file')
        response = self.client.post(reverse('api:file-upload', kwargs={'pk': self.corpus.id}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        f = SimpleUploadedFile('test2.txt', b'This is a text file')
        response = self.client.post(reverse('api:file-upload', kwargs={'pk': self.corpus.id}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(s3_mock.Object.call_count, 1)
        self.assertEqual(s3_mock.Object().upload_fileobj.call_count, 1)

    def test_file_upload_demo_limit(self):
        """
        Assert a file upload fails for demo users when its size exceeds the corpus limit
        """
        self.client.force_login(self.demo_user)
        # Create a file one byte under the allowed size
        DataFile.objects.create(
            name='test.txt', size=52428799, hash='aaaa', content_type='text/plain', corpus=self.corpus)
        # Another file of morethan one byte in size
        f = SimpleUploadedFile('test.txt', b'This is a text file')
        response = self.client.post(reverse('api:file-upload', kwargs={'pk': self.corpus.id}), data={'file': f})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
