from unittest.mock import patch, call
from arkindex.project.tests import FixtureTestCase
from arkindex.documents.models import ElementType
from arkindex.dataimport.tasks import check_images, build_iiif_image
from arkindex.images.models import ImageStatus
from botocore.exceptions import ClientError


class TestImageTasks(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.vol = cls.corpus.elements.create(name='Test volume', type=ElementType.Volume)
        cls.df = cls.corpus.files.create(
            name='file.jpg',
            size=1234,
            hash='cafe',
            content_type='image/jpeg',
        )

    @patch('arkindex.dataimport.tasks.image.PillowImage')
    def test_check_images(self, image_mock):
        image_mock.open.return_value.size = (1000, 1000)

        result = check_images([(self.df, '/some/path'), ])

        self.assertEqual(image_mock.open.call_count, 1)
        self.assertEqual(image_mock.open.call_args, call('/some/path'))
        self.assertListEqual(result, [
            (self.df, '/some/path'),
        ])

    @patch('arkindex.images.models.s3.Object')
    @patch('arkindex.dataimport.tasks.image.PillowImage')
    def test_build_iiif_image(self, image_mock, s3obj_mock):
        image_mock.open.return_value.format = 'BMP'
        image_mock.open.return_value.size = (400, 900)
        s3obj_mock.return_value.load.side_effect = ClientError({'Error': {'Code': '404'}}, 'head_object')

        with self.settings(LOCAL_IMAGESERVER_ID=self.imgsrv.id, AWS_IIIF_BUCKET='iiif'):
            img = build_iiif_image(
                self.vol,
                '/somewhere/Untitled.bmp',
                self.df,
                suffix='42',
            )

        expected_path = '{}/{}-42.jpg'.format(str(self.vol.id), str(self.df.id))
        self.assertEqual(image_mock.open.call_count, 1)
        self.assertEqual(image_mock.open.call_args, call('/somewhere/Untitled.bmp'))
        self.assertEqual(s3obj_mock.call_count, 1)
        self.assertEqual(s3obj_mock.call_args, call('iiif', expected_path))
        self.assertEqual(s3obj_mock().load.call_count, 1)
        self.assertEqual(image_mock.open().save.call_count, 1)
        self.assertEqual(s3obj_mock().upload_fileobj.call_count, 1)
        self.assertEqual(img.server, self.imgsrv)
        self.assertEqual(img.path, expected_path.replace('/', '%2F'))
        self.assertEqual(img.width, 400)
        self.assertEqual(img.height, 900)
        self.assertEqual(img.status, ImageStatus.Checked)
        self.assertEqual(img.datafile, self.df)

    @patch('arkindex.images.models.s3.Object')
    @patch('arkindex.dataimport.tasks.image.PillowImage')
    def test_build_iiif_image_retry(self, image_mock, s3obj_mock):
        """
        Test build_iiif_image just returns existing images if they already exist
        """
        image_mock.open.return_value.format = 'JPEG2000'
        image_mock.open.return_value.size = (400, 900)
        original_img = self.imgsrv.images.create(
            path='{}%2F{}.jp2'.format(str(self.vol.id), str(self.df.id)),
            datafile=self.df,
            width=900,
            height=400,
            status=ImageStatus.Checked,
        )

        with self.settings(LOCAL_IMAGESERVER_ID=self.imgsrv.id, AWS_IIIF_BUCKET='iiif'):
            new_img = build_iiif_image(
                self.vol,
                '/somewhere/Untitled.bmp',
                self.df,
            )

        self.assertEqual(original_img.id, new_img.id)
        self.assertEqual(image_mock.open.call_count, 1)
        self.assertEqual(image_mock.open.call_args, call('/somewhere/Untitled.bmp'))
        self.assertEqual(s3obj_mock().load.call_count, 1)
        self.assertFalse(image_mock.open().save.called)
