from rest_framework import status
from django.urls import reverse
from arkindex.dataimport.models import \
    DataImport, DataImportMode, DataFile
from arkindex.documents.models import Element, ElementType, Corpus
from arkindex.project.tests import FixtureAPITestCase
from ponos.models import State
import uuid


class TestImports(FixtureAPITestCase):
    """
    Test data imports management
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get()
        cls.rev = cls.repo.revisions.get()
        cls.demo_volume = Element.objects.create(type=ElementType.Volume, name='Demo', corpus=cls.corpus)
        cls.img_df = DataFile.objects.create(
            name='test.jpg', size=42, hash='aaaa', content_type='image/jpg', corpus=cls.corpus)
        cls.pdf_df = DataFile.objects.create(
            name='test.pdf', size=42, hash='bbbb', content_type='application/pdf', corpus=cls.corpus)

    def setUp(self):
        super().setUp()
        self.dataimport = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Images)

    def test_list_requires_login(self):
        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        result = data['results'][0]
        self.assertEqual(result['id'], str(self.dataimport.id))
        self.assertEqual(result['state'], State.Unscheduled.value)
        self.assertEqual(result['mode'], DataImportMode.Images.value)
        self.assertEqual(result['corpus'], str(self.corpus.id))

    def test_list_filter_corpus(self):
        self.client.force_login(self.superuser)
        corpus2 = Corpus.objects.create(name='Another corpus', description='something')
        dataimport2 = corpus2.imports.create(creator=self.user, mode=DataImportMode.Images)

        response = self.client.get(reverse('api:import-list'), {'corpus': str(corpus2.id)})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(dataimport2.id))

    def test_list_filter_mode(self):
        self.client.force_login(self.user)
        dataimport2 = self.corpus.imports.create(
            creator=self.user,
            mode=DataImportMode.PDF,
        )
        response = self.client.get(reverse('api:import-list'), {'mode': DataImportMode.PDF.value})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(dataimport2.id))

    def test_list_filter_unexisting_mode(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'mode': 'unexisting_mode'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_list_filter_id(self):
        self.client.force_login(self.user)

        # Generate a UUID whose first characters are not the same as the first DataImport
        dataimport_id = str(uuid.uuid4())
        while dataimport_id[:10] == str(self.dataimport.id)[:10]:
            dataimport_id = str(uuid.uuid4())

        dataimport2 = self.corpus.imports.create(
            id=dataimport_id,
            creator=self.user,
            mode=DataImportMode.Images,
        )
        response = self.client.get(reverse('api:import-list'), {'id': dataimport_id[:10]})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(dataimport2.id))

    def test_details_requires_login(self):
        response = self.client.get(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_details_destroy_no_permission(self):
        self.client.force_login(self.user)
        acl_rights = self.corpus.corpus_right.get(user=self.user)
        acl_rights.can_write = False
        acl_rights.can_admin = False
        acl_rights.save()
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_details_destroy_running(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        task = self.dataimport.workflow.tasks.all().last()
        task.state = State.Running
        task.save()
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_details_destroy(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_details(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.dataimport.id))

    def test_start_demo_requires_login(self):
        response = self.client.post(reverse('api:import-demo', kwargs={'pk': self.img_df.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_start_demo(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-demo', kwargs={'pk': self.img_df.id}))
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.Images)
        self.assertEqual(dataimport.workflow.state, State.Unscheduled)
        self.assertListEqual(list(dataimport.files.all()), [self.img_df])
        self.assertEqual(dataimport.payload['volume_id'], str(self.demo_volume.id))
        self.assertEqual(dataimport.payload['folder_name'], 'user-{}-demo'.format(self.user.id))

    def test_retry_requires_login(self):
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retry_only_final(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        self.assertEqual(self.dataimport.state, State.Unscheduled)
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_retry_running(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        task = self.dataimport.workflow.tasks.all().last()
        task.state = State.Running
        task.save()
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_retry(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(self.dataimport.state, State.Error)
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataimport.refresh_from_db()
        self.assertEqual(self.dataimport.state, State.Unscheduled)

    def test_retry_repo_disabled(self):
        self.client.force_login(self.user)
        self.dataimport.mode = DataImportMode.Repository
        self.dataimport.revision = self.rev
        self.dataimport.save()
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(self.dataimport.state, State.Error)
        self.creds.delete()
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.creds.save()

    def test_failures_requires_login(self):
        self.dataimport.failures.create(path='path/to/file', message='something')
        response = self.client.get(reverse('api:import-failures', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_failures(self):
        self.client.force_login(self.user)
        failure = self.dataimport.failures.create(path='path/to/file', message='something')
        response = self.client.get(reverse('api:import-failures', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(failure.id))

    def test_from_files_requires_login(self):
        response = self.client.post(reverse('api:import-from-files'), {'files': [str(self.img_df.id)]}, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_from_files(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {'files': [str(self.img_df.id)]}, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.Images)
        self.assertListEqual(list(dataimport.files.all()), [self.img_df])
        self.assertIn('volume_id', dataimport.payload)
        self.assertIn('folder_name', dataimport.payload)

    def test_from_files_pdf(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:import-from-files'),
            {'files': [str(self.pdf_df.id)], 'mode': 'pdf'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.PDF)
        self.assertListEqual(list(dataimport.files.all()), [self.pdf_df])
        self.assertIn('volume_id', dataimport.payload)
        self.assertIn('folder_name', dataimport.payload)
        self.assertEqual(Element.objects.get(id=dataimport.payload['volume_id']).name, 'Import pdf test')

    def test_from_files_invalid_mode(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:import-from-files'),
            {'files': [str(self.pdf_df.id)], 'mode': 'repository'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_pdf_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:import-from-files'),
            {'files': [str(self.img_df.id)], 'mode': 'pdf'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_pdf_single(self):
        self.client.force_login(self.user)
        pdf2 = DataFile.objects.create(
            name='test2.pdf', size=1337, hash='0', content_type='application/pdf', corpus=self.corpus)
        response = self.client.post(
            reverse('api:import-from-files'),
            {'files': [str(self.pdf_df.id), str(pdf2.id)], 'mode': 'pdf'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_images_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:import-from-files'),
            {'files': [str(self.pdf_df.id)]},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_volume_id(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:import-from-files'),
            {
                'files': [str(self.pdf_df.id)],
                'mode': 'pdf',
                'volume_id': str(self.demo_volume.id),
            },
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.payload['volume_id'], str(self.demo_volume.id))

    def test_from_files_volume_name(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:import-from-files'),
            {
                'files': [str(self.pdf_df.id)],
                'mode': 'pdf',
                'volume_name': 'New volume',
            },
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        volume = Element.objects.get(id=dataimport.payload['volume_id'])
        self.assertEqual(volume.name, 'New volume')
        self.assertEqual(volume.corpus, self.corpus)

    def test_from_files_no_files(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {'files': []}, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
