from arkindex.project.tests import FixtureTestCase
from arkindex.dataimport.tasks import extract_pdf_images
from unittest.mock import patch, MagicMock
from botocore.exceptions import ClientError
import tempfile
import shutil
import os.path

FIXTURES = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    'pdf_samples',
)


class TestPdf(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.img_file = cls.corpus.files.create(name='sample.jpg', size=42, hash='abcd', content_type='image/jpg')
        cls.pdf_file = cls.corpus.files.create(name='sample.pdf', size=42, hash='dcba', content_type='application/pdf')

        cls.working_dir = tempfile.mkdtemp()
        cls.pdf_path = os.path.join(cls.working_dir, str(cls.pdf_file.id))
        shutil.copyfile(os.path.join(FIXTURES, 'sample.pdf'), cls.pdf_path)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        shutil.rmtree(cls.working_dir)

    def test_extract_pdf_images_filetype(self):
        """
        Test extract_pdf_images task the file's content type
        """
        with self.assertRaises(AssertionError):
            extract_pdf_images(self.img_file, self.pdf_path, self.working_dir)

    def test_extract_pdf_images_exists(self):
        """
        Test extract_pdf_images task checks the file's existence
        """
        file_mock = MagicMock()
        file_mock.content_type = 'application/pdf'
        file_mock.exists.return_value = False

        with self.assertRaises(AssertionError):
            extract_pdf_images(file_mock, self.pdf_path, self.working_dir)

    @patch('arkindex.dataimport.models.s3')
    def test_extract_pdf_images_s3_error(self, s3_mock):
        """
        Test extract_pdf_images task lets S3 errors through
        """
        file_mock = MagicMock()
        file_mock.content_type = 'application/pdf'
        # Any ClientError with a code other than 404
        file_mock.exists.side_effect = ClientError({'Error': {'Code': '999'}}, 'head_object')

        with self.assertRaises(ClientError):
            extract_pdf_images(file_mock, self.pdf_path, self.working_dir)

    @patch('arkindex.dataimport.models.s3')
    def test_extract_pdf_images(self, s3_mock):
        """
        Test extract_pdf_images runs ImageMagick and returns proper info
        """
        result = extract_pdf_images(self.pdf_file, self.pdf_path, self.working_dir)

        self.assertListEqual(result, [
            os.path.join(self.working_dir, 'pdf-0000.jpg'),
            os.path.join(self.working_dir, 'pdf-0001.jpg'),
        ])
