from arkindex.project.tests import FixtureTestCase
from arkindex.dataimport.models import DataImportMode
from arkindex.dataimport.tasks import download_files
from unittest.mock import patch, call


class TestTasks(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.df1 = cls.corpus.files.create(
            name='file1.jpg',
            size=1234,
            hash='cafe',
            content_type='image/jpeg',
        )
        cls.df2 = cls.corpus.files.create(
            name='file2.png',
            size=5678,
            hash='beef',
            content_type='image/png',
        )
        cls.di = cls.corpus.imports.create(
            mode=DataImportMode.Images,
            creator=cls.user,
        )

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.s3obj_patch = patch('arkindex.dataimport.models.s3.Object')
        cls.s3obj_mock = cls.s3obj_patch.start()
        cls.access_patch = patch('arkindex.dataimport.tasks.base.os.access')
        cls.access_mock = cls.access_patch.start()

    @classmethod
    def tearDownClass(cls):
        cls.s3obj_patch.stop()
        cls.access_patch.stop()
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.s3obj_mock.reset_mock()
        self.access_mock.reset_mock()
        self.access_mock.return_value = True
        self.s3obj_mock.return_value.download_file.side_effect = None

    def test_download_files(self):
        self.di.files.set([self.df1])
        expected_path = '/somewhere/{}'.format(str(self.df1.id))
        self.assertListEqual(
            download_files(self.di, '/somewhere'),
            [(self.df1, expected_path)],
        )
        self.assertEqual(self.s3obj_mock().download_file.call_count, 1)
        self.assertEqual(self.s3obj_mock().download_file.call_args, call(expected_path))

    def test_download_files_fail(self):
        self.di.files.set([self.df1, self.df2])
        # Fail only once
        self.s3obj_mock().download_file.side_effect = [IOError, None]

        expected_path1 = '/somewhere/{}'.format(str(self.df1.id))
        expected_path2 = '/somewhere/{}'.format(str(self.df2.id))

        self.assertListEqual(
            download_files(self.di, '/somewhere'),
            [(self.df2, expected_path2)],
        )

        self.assertEqual(self.s3obj_mock().download_file.call_count, 2)
        self.assertEqual(self.s3obj_mock().download_file.call_args_list, [
            call(expected_path1),
            call(expected_path2),
        ])

    def test_download_files_epic_fail(self):
        self.di.files.set([self.df1, self.df2])
        # Fail all the time
        self.s3obj_mock().download_file.side_effect = IOError

        expected_path1 = '/somewhere/{}'.format(str(self.df1.id))
        expected_path2 = '/somewhere/{}'.format(str(self.df2.id))

        with self.assertRaisesRegex(AssertionError, 'No files'):
            download_files(self.di, '/somewhere')

        self.assertEqual(self.s3obj_mock().download_file.call_count, 2)
        self.assertEqual(self.s3obj_mock().download_file.call_args_list, [
            call(expected_path1),
            call(expected_path2),
        ])

    def test_download_files_empty(self):
        self.di.files.set([])
        self.assertListEqual(download_files(self.di, '/somewhere'), [])
        self.assertFalse(self.s3obj_mock.called)

    def test_download_files_read_only(self):
        self.access_mock.return_value = False
        self.di.files.set([self.df1])
        with self.assertRaisesRegex(AssertionError, 'read-only'):
            download_files(self.di, '/somewhere')
