from django.contrib import admin
from django.urls import path, reverse
from django.utils.html import format_html
from arkindex.documents.models import \
    Corpus, Page, Element, ElementType, Act, Transcription, MetaData, InterpretedDate, Classification, DataSource
from arkindex.documents.views import DumpActs
from arkindex.dataimport.models import Event
from enumfields.admin import EnumFieldListFilter


class CorpusAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'public', )


class DataSourceAdmin(admin.ModelAdmin):
    list_display = ('id', 'type', 'slug', 'revision', 'internal')
    list_filter = [('type', EnumFieldListFilter), 'internal']
    readonly_fields = ('id', )


class EventInline(admin.TabularInline):
    model = Event


class ClassificationInline(admin.TabularInline):
    model = Classification


class PageAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'page_type', 'nb', 'direction', )
    list_filter = [('page_type', EnumFieldListFilter), 'corpus']
    fields = ('id', 'name', 'folio', 'page_type', 'nb', 'direction')
    readonly_fields = ('id', )
    search_fields = ('name', )
    inlines = (EventInline, ClassificationInline, )


class DateInline(admin.TabularInline):
    extra = 1
    model = InterpretedDate


class MetaDataAdmin(admin.ModelAdmin):
    list_display = ('id', 'type', 'revision')
    readonly_fields = ('id', 'revision')
    raw_id_fields = ('element', )
    inlines = (DateInline, )


class MetaDataInline(admin.TabularInline):
    model = MetaData


class ElementAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'type', 'corpus', 'element_actions')
    list_filter = [('type', EnumFieldListFilter), 'corpus']
    fields = ('id', 'type', 'name', 'corpus')
    readonly_fields = ('id', 'element_actions')
    search_fields = ('name', )
    inlines = (MetaDataInline, EventInline)

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path(
                '<uuid:pk>/acts/',
                self.admin_site.admin_view(DumpActs.as_view()),
                name='dump-acts',
            ),
        ]
        return custom_urls + urls

    def element_actions(self, obj):
        return format_html(
            '<a class="button" href="{}">Dump acts</a>',
            reverse('admin:dump-acts', args=[obj.pk]),
        ) if obj.type == ElementType.Volume else ''
    element_actions.short_description = 'Actions'
    element_actions.allow_tags = True


class ActAdmin(admin.ModelAdmin):
    list_display = ('id', 'name')
    fields = ('id', 'name', 'folio', 'number')
    readonly_fields = ('id', )
    search_fields = ('name', )
    inlines = (MetaDataInline, EventInline)


class TranscriptionAdmin(admin.ModelAdmin):
    list_display = ('id', 'text', 'score', 'element', )
    list_filter = [('type', EnumFieldListFilter), 'source']
    fields = ('id', 'text', 'score', 'element', 'zone', 'source', )
    readonly_fields = ('id', )
    raw_id_fields = ('element', 'zone', )


admin.site.register(Corpus, CorpusAdmin)
admin.site.register(DataSource, DataSourceAdmin)
admin.site.register(Page, PageAdmin)
admin.site.register(Element, ElementAdmin)
admin.site.register(Act, ActAdmin)
admin.site.register(Transcription, TranscriptionAdmin)
admin.site.register(MetaData, MetaDataAdmin)
