from django.core.exceptions import ValidationError
from arkindex.documents.models import DateType, InterpretedDate
import unicodedata
import re
import logging

logger = logging.getLogger(__name__)

MONTHS = {
    'en': (
        'january',
        'february',
        'march',
        'april',
        'may',
        'june',
        'july',
        'august',
        'september',
        'october',
        'november',
        'december',
        ),
    'fr': (
        'janvier',
        'fevrier',
        'mars',
        'avril',
        'mai',
        'juin',
        'juillet',
        'aout',
        'septembre',
        'octobre',
        'novembre',
        'decembre',
    ),
}


def year(raw_date):
    """
    Matches 4-digit year only: '1320', '2010'
    """
    match = re.match(r'^(?P<year>\d{4})$', raw_date)
    return match.groupdict() if match else None


def year_month(raw_date):
    """
    Matches 4-digit year followed by 1 or 2-digit month: '1320-3', '2010-02'
    """
    match = re.match(r'^(?P<year>\d{4})-(?P<month>\d{1,2})$', raw_date)
    return match.groupdict() if match else None


def year_month_day(raw_date):
    """
    Matches 4-digit year followed by 1 or 2-digit month then
    1 or 2-digit day : '1320-3-12', '2010-02-31'
    """
    match = re.match(r'^(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})$', raw_date)
    return match.groupdict() if match else None


def year_month_str(raw_date):
    """
    Matches a date with alpha-written month and 4-digit year
    Both must be unique : '1320, march', 'february 2010'
    Days are not interpreted yet if present in the raw date text
    """
    years = []
    months = []
    for match in re.finditer(r'(?=(^|[^\d])(?P<year>\d{4})($|\D))', raw_date):
        years.append(match.groupdict()['year'])
    words = re.findall(r'[a-z]+', raw_date)
    for w in words:
        for month_lang in MONTHS:
            if w in MONTHS[month_lang]:
                months.append(MONTHS[month_lang].index(w) + 1)
                break
    # Exactly one year and 1 month are found
    if len(years) == len(months) == 1:
        return {'year': years[0], 'month': months[0]}
    return None


class DateParser(object):
    """
    Return interpreted date elements from non-standard TEI input
    Input examples :  "1325-11-04"
                      "s.d. [1323-1325]"
    """

    date_table = (
        year,
        year_month,
        year_month_day,
        year_month_str,
    )

    def _normalize(self, raw_date):
        """
        Return an unaccented lowercase string from input
        """
        return unicodedata.normalize('NFKD', raw_date) \
            .encode('ASCII', 'ignore') \
            .lower() \
            .decode("utf-8")

    def parse(self, raw_date):
        assert isinstance(raw_date, str)
        # Try to match regex one by one
        for f in self.date_table:
            try:
                date_elts = f(self._normalize(raw_date))
                if date_elts:
                    assert isinstance(date_elts, dict)
                    date = InterpretedDate(type=DateType.Exact)
                    for k in date_elts:
                        setattr(date, k, int(date_elts[k]))
                    try:
                        date.full_clean(exclude=('metadata', ))
                    except ValidationError as e:
                        logger.warning('Date fields are incorrect : {}'.format(e))
                        continue
                    return (date, )
            except Exception as e:
                logger.warning('Exception {} has been raised during parsing of date {}'.format(e, raw_date), f)
        logger.warning('Date not supported : {}'.format(raw_date))
        return ()
