from django.conf import settings
from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk as es_bulk
from elasticsearch.exceptions import NotFoundError, RequestError
from arkindex.documents.models import Act, Page
import logging
import time
import datetime

logger = logging.getLogger(__name__)


class Indexer(object):

    def __init__(self, hosts=settings.ELASTIC_SEARCH_HOSTS):
        self.elastic = Elasticsearch(hosts)

    def setup(self):
        """
        Create indexes in ElasticSearch
        """
        try:
            self.elastic.indices.create(
                index=settings.ES_INDEX_ACTS,
                body={
                    "mappings": {
                        Act.INDEX_TYPE: {
                            "properties": {
                                "corpus": {"type": "keyword"},
                                "volumes": {"type": "keyword"},
                                "transcriptions": {
                                    "type": "nested",
                                    "properties": {
                                        "id": {"type": "keyword"},
                                        "type": {"type": "keyword"},
                                        "score": {"type": "float"},
                                        "text": {"type": "text"},
                                    }
                                },
                                "date_range": {
                                    "type": "date_range",
                                    "format": "yyyy||yyyy-MM||yyyy-MM-dd"
                                }
                            }
                        }
                    }
                }
            )
        except RequestError:  # Index already exists
            pass
        try:
            self.elastic.indices.create(
                index=settings.ES_INDEX_PAGES,
                body={
                    "mappings": {
                        Page.INDEX_TYPE: {
                            "properties": {
                                "corpus": {"type": "keyword"},
                                "volumes": {"type": "keyword"},
                                "transcriptions": {
                                    "type": "nested",
                                    "properties": {
                                        "id": {"type": "keyword"},
                                        "type": {"type": "keyword"},
                                        "score": {"type": "float"},
                                        "text": {"type": "text"},
                                    }
                                }
                            }
                        }
                    }
                }
            )
        except RequestError:  # Index already exists
            pass

    def drop_index(self, index_name):
        """
        Try to drop an existing index
        """
        try:
            self.elastic.indices.delete(index=index_name)
            logger.info("Dropped index {}".format(index_name))
        except NotFoundError:
            logger.info("Could not drop index {} (does not exist)".format(index_name))

    def run_index(self, index_name, index_type, queryset, bulk_size=400):
        """
        Reindex elements from a QuerySet chunk by chunk and output status
        """
        count = queryset.count()
        logger.info("{} items to reindex".format(count))

        start_time, last, indexed = time.time(), 0, 0
        for i in range(0, count, bulk_size):
            percent = int(100.0 * i / count)
            if last != percent or bulk_size > 100:
                secs = time.time() - start_time
                per_second = int(i / secs)
                logger.info(
                    '{0: >3}% => {1: >5} / {2} (indexed: {3}, execution time: {4}, {5} items/s)'.format(
                        percent, i, count, indexed, datetime.timedelta(seconds=secs), per_second))
                last = percent

            indexed += self.index(index_name, index_type, queryset[i:i+bulk_size])

    def index(self, index_name, index_type, queryset, retries=3, chunk_size=100, timeout=15):
        """
        Insert or update items into ElasticSearch
        """
        assert retries > 0

        # Build raw ElasticSearch insert
        actions = [{
            '_index': index_name,
            '_type': index_type,
            '_source': item.build_search_index(),
            '_id': item.id.hex,
        } for item in queryset.all()]

        # Run actions in bulk
        try:
            nb_insert, _ = es_bulk(
                self.elastic,
                actions,
                chunk_size=chunk_size,
                initial_backoff=2,
                max_retries=2,
                request_timeout=timeout,
                stats_only=True,
                _source=False,
            )
        except Exception as e:
            if retries <= 1:
                logger.warning('Failed to bulk insert into ES : {}'.format(e))
                return 0
            logger.warning('Failed to bulk insert into ES - retrying : {}'.format(e))
            return self.index(
                index_name,
                index_type,
                queryset,
                retries=retries-1,
                chunk_size=max(chunk_size // 2, 1),
                timeout=timeout*2,
            )

        return nb_insert
