#!/usr/bin/env python3
from django.conf import settings
from django.db import transaction
from ponos.management.base import PonosCommand
from arkindex.project.argparse import CorpusArgument
from arkindex.documents.models import MetaData, MetaType
from arkindex.documents.date_parser import DateParser
import logging

logging.basicConfig(
    format='%(asctime)s [%(levelname)s] %(message)s',
    level=logging.INFO,
)

logger = logging.getLogger(__name__)


class Command(PonosCommand):
    help = "Create fresh interpreted dates from metadatas"
    docker_image = settings.ARKINDEX_APP_IMAGE
    base_recipe = settings.PONOS_RECIPE

    def add_arguments(self, parser):
        parser.add_argument(
            '--corpus',
            help='ID or part of the name of the corpus to fetch metadatas from',
            type=CorpusArgument(),
        )
        parser.add_argument(
            '--dry-run',
            action='store_true',
            default=False,
            help='Print number of dates to update',
        )

    def run(self, corpus=None, dry_run=False, **options):
        metadatas = MetaData.objects.filter(type=MetaType.Date)
        if corpus:
            metadatas = metadatas.filter(element__corpus_id=corpus.id)
        total_metadatas = len(metadatas)
        covered_metadatas = sum(1 if m.dates.count() else 0 for m in metadatas)
        logger.info('Found {} MetaData elements corresponding to a Date'.format(total_metadatas))
        date_parser = DateParser()
        covered_metadatas_update = 0
        if not dry_run:
            deleted_dates_num = 0
            with transaction.atomic():
                for metadata in metadatas:
                    deleted = metadata.dates.all().delete()
                    logger.debug('Deleted {} interpreted date(s) for {} Metadata'.format(deleted[0], metadata.value))
                    deleted_dates_num += deleted[0]
            logger.info('{} old InterpretedDates deleted from database'.format(deleted_dates_num))
        for metadata in metadatas:
            updated = 0
            for date in date_parser.parse(metadata.value):
                date.metadata = metadata
                updated = 1
                if not dry_run:
                    date.save()
            covered_metadatas_update += updated
        logger.info('{} values have been updated'.format(covered_metadatas_update))
        logger.info('Updated coverage is {:.2%} ({:+.2%})'.format(
            covered_metadatas_update / total_metadatas,
            (covered_metadatas_update - covered_metadatas) / total_metadatas,
        ))
