#!/usr/bin/env python3
from django.conf import settings
from ponos.management.base import PonosCommand
from arkindex.project.argparse import ElementArgument
from arkindex.documents.indexer import Indexer
from arkindex.documents.models import Element, ElementType, Act, Transcription, Page
import logging

logging.basicConfig(
    level=logging.INFO,
    format='[%(levelname)s] %(message)s',
)
logger = logging.getLogger(__name__)


def get_acts(volume=None):
    if volume:
        queryset = Act.objects.get_descending(volume.id)
    else:
        queryset = Act.objects.all()

    return queryset.prefetch_related('metadatas__dates')


def get_transcriptions(volume=None):
    if volume:
        # Lookup all the transcriptions linked to a volume
        return Transcription.objects.filter(
            element__in=Element.objects.get_descending(volume.id)
        ).distinct()

    return Transcription.objects.all()


def get_pages(volume=None):
    if volume:
        return Page.objects.get_descending(volume.id)

    return Page.objects.filter(
        id__in=Transcription.objects.values_list('element', flat=True).distinct(),
    )


class Command(PonosCommand):
    help = 'Selectively reindex transcriptions, pages and acts into ElasticSearch'
    docker_image = settings.ARKINDEX_APP_IMAGE
    base_recipe = settings.PONOS_RECIPE

    # Setup for reindexation of different Elements
    index_methods = {
        'acts': {
            'bulk_size': 100,
            'index': settings.ES_INDEX_ACTS,
            'type': Act.INDEX_TYPE,
            'items': get_acts,
        },
        'pages': {
            'bulk_size': 100,
            'index': settings.ES_INDEX_PAGES,
            'type': Page.INDEX_TYPE,
            'items': get_pages,
        },
        'transcriptions': {
            'bulk_size': 400,
            'index': settings.ES_INDEX_TRANSCRIPTIONS,
            'type': Transcription.INDEX_TYPE,
            'items': get_transcriptions,
        },
    }

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            '--acts',
            help='Reindex acts',
            action='store_true',
        )
        parser.add_argument(
            '-ts', '--transcriptions',
            help='Reindex transcriptions',
            action='store_true',
        )
        parser.add_argument(
            '--pages',
            help='Reindex pages',
            action='store_true',
        )
        parser.add_argument(
            '--volume',
            help='Restrict reindexing to a specific volume by ID or part of the name',
            type=ElementArgument(type=ElementType.Volume),
        )
        parser.add_argument(
            '--drop',
            help="Drop the existing indexes before reindexing",
            action='store_true',
        )

    def validate_args(self, **options):
        # If no specific index was set, reindex everything
        if not any(options.get(k) for k in self.index_methods.keys()):
            for k in self.index_methods.keys():
                options[k] = True

        return {
            'methods': [
                key
                for key in self.index_methods.keys()
                if options.get(key)
            ],
            'volume': options['volume'],
            'drop': options.get('drop', False),
        }

    def run(self, methods=[], drop=False, volume=None):
        '''
        Index all prepared elements
        '''
        indexer = Indexer()
        for method in methods:
            config = self.index_methods[method]

            if drop:
                indexer.drop_index(config['index'])
                indexer.setup()

            items = config['items'](volume)

            indexer.run_index(
                config['index'],
                config['type'],
                items,
                bulk_size=config['bulk_size'],
            )
