from rest_framework import serializers
from arkindex.documents.models import \
    Element, ElementType, Page, PageType, PageDirection, Act, Corpus, \
    MetaData, MetaType, InterpretedDate, DateType
from arkindex.images.serializers import ZoneSerializer, ImageSerializer
from arkindex.documents.serializers.light import ElementLightSerializer, CorpusLightSerializer
from arkindex.documents.serializers.ml import ClassificationSerializer, TranscriptionSerializer
from arkindex.dataimport.serializers import RevisionSerializer
from arkindex.dataimport.models import EventType
from arkindex.project.serializer_fields import EnumField


class InterpretedDateSerializer(serializers.ModelSerializer):
    """
    Serialize a list of interpreted dates linked to a metadata
    """
    type = EnumField(DateType)

    class Meta:
        model = InterpretedDate
        fields = (
            'year',
            'month',
            'day',
            'type',
        )


class MetaDataSerializer(serializers.ModelSerializer):
    """
    Serialises some Metadata for any Element
    """
    type = EnumField(MetaType)
    revision = RevisionSerializer()
    dates = InterpretedDateSerializer(many=True)

    class Meta:
        model = MetaData
        fields = (
            'id',
            'type',
            'name',
            'value',
            'revision',
            'dates',
        )


class PageSerializer(serializers.ModelSerializer):
    """
    Serialises a Page
    """
    page_type = EnumField(PageType)
    direction = EnumField(PageDirection)
    image = ImageSerializer(source='zone.image')
    corpus = CorpusLightSerializer()

    parents = serializers.ListField(
        child=serializers.ListField(
            child=ElementLightSerializer(),
        ),
        source='parent_elements',
        read_only=True,
    )

    transcriptions = TranscriptionSerializer(
        source='page_transcriptions',
        many=True,
    )
    classifications = ClassificationSerializer(many=True)

    class Meta:
        model = Page
        fields = (
            'id',
            'page_type',
            'nb',
            'direction',
            'display_name',
            'image',
            'corpus',
            'parents',
            'transcriptions',
            'classifications',
        )


class CorpusSerializer(serializers.ModelSerializer):
    """
    Serialize a corpus
    """
    rights = serializers.SerializerMethodField(read_only=True)

    class Meta:
        model = Corpus
        fields = (
            'id',
            'name',
            'description',
            'public',
            'rights',
        )

    def get_rights(self, corpus):
        return map(lambda x: x.value, corpus.get_acl_rights(self.context['request'].user))

    def create(self, validated_data):
        c = Corpus.objects.create(**validated_data)
        c.corpus_right.create(
            user=self.context['request'].user,
            can_write=True,
            can_admin=True,
        )
        return c


class ElementLightSerializer(serializers.ModelSerializer):
    """
    Fully Serialises a document
    """
    type = EnumField(ElementType, read_only=True)
    corpus = CorpusLightSerializer(read_only=True)
    thumbnail_url = serializers.URLField(source='thumbnail.url', read_only=True)

    class Meta:
        model = Element
        fields = (
            'id',
            'type',
            'name',
            'corpus',
            'thumbnail_url',
        )


class ElementSerializer(ElementLightSerializer):
    """
    Serialize an element with its metadata
    """

    metadata = serializers.SerializerMethodField()

    def get_metadata(self, obj):
        return MetaDataSerializer(sorted(list(obj.metadatas.all())), many=True).data

    class Meta:
        model = Element
        fields = ElementLightSerializer.Meta.fields + (
            'metadata',
        )

    def create(self, *args, **kwargs):
        instance = super().create(*args, **kwargs)
        instance.events.create(type=EventType.Addition)
        return instance

    def update(self, *args, **kwargs):
        instance = super().update(*args, **kwargs)
        instance.events.create(type=EventType.Edit)
        return instance


class ActSerializer(serializers.ModelSerializer):
    """
    Serialize an act with its parents and children
    and metadatas
    """

    parents = serializers.ListField(
        child=serializers.ListField(
            child=ElementLightSerializer()
        ),
        source='parent_elements',
        read_only=True,
    )
    children = ElementLightSerializer(source='child_elements', read_only=True, many=True)
    metadata = serializers.SerializerMethodField()

    def get_metadata(self, obj):
        return MetaDataSerializer(sorted(list(obj.metadatas.all())), many=True).data

    class Meta:
        model = Act
        fields = (
            'id',
            'name',
            'number',
            'parents',
            'children',
            'metadata',
        )
        read_only_fields = ('id', )


class SurfaceSerializer(serializers.ModelSerializer):
    """
    Serialize a surface with its page, image and zone
    """

    zone = ZoneSerializer()
    page = PageSerializer()

    class Meta:
        model = Element
        fields = (
            'id',
            'name',
            'page',
            'zone',
        )
