from rest_framework import serializers
from arkindex_common.ml_tool import MLToolType
from arkindex.project.serializer_fields import EnumField, MLToolField
from arkindex.documents.models import \
    Corpus, Element, Page, Transcription, TranscriptionType, DataSource, Classification
from arkindex.images.serializers import ZoneSerializer


class DataSourceSerializer(serializers.ModelSerializer):
    """
    Serialize a data source for transcriptions and classifications
    """

    type = EnumField(MLToolType)

    class Meta:
        model = DataSource
        fields = (
            'id',
            'type',
            'slug',
            'revision',
            'internal',
        )


class ClassificationSerializer(serializers.ModelSerializer):
    """
    Serialize a classification on a Page
    """

    source = DataSourceSerializer()

    class Meta:
        model = Classification
        fields = (
            'source',
            'class_name',
            'confidence',
        )


class TranscriptionSerializer(serializers.ModelSerializer):
    """
    Serialises a Transcription
    issued from a search
    """

    type = EnumField(TranscriptionType)
    zone = ZoneSerializer()
    source = DataSourceSerializer()

    class Meta:
        model = Transcription
        fields = (
            'id',
            'type',
            'text',
            'score',
            'zone',
            'source',
        )


class TranscriptionCreateSerializer(serializers.Serializer):
    """
    Allows for insertion of new transcriptions and zones
    """
    element = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.filter(zone__isnull=False),
    )
    source = serializers.PrimaryKeyRelatedField(
        queryset=DataSource.objects.filter(type=MLToolType.Recognizer),
    )
    polygon = serializers.ListField(
        child=serializers.ListField(
            child=serializers.IntegerField(),
            min_length=2,
            max_length=2
        ),
        min_length=3
    )
    text = serializers.CharField()
    score = serializers.FloatField(min_value=0, max_value=1)
    type = EnumField(TranscriptionType)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):  # May be None when generating an OpenAPI schema or using from a REPL
            return
        self.fields['element'].queryset = Element.objects.filter(
            zone__isnull=False,
            corpus__in=Corpus.objects.writable(self.context['request'].user),
        )


class TranscriptionBulkSerializer(serializers.Serializer):
    """
    Allows for insertion of new transcriptions and zones
    in Bulk (used by serializer below)
    Note: no element !
    """
    polygon = serializers.ListField(
        child=serializers.ListField(
            child=serializers.IntegerField(),
            min_length=2,
            max_length=2
        ),
        min_length=3
    )
    text = serializers.CharField()
    score = serializers.FloatField(min_value=0, max_value=1)
    type = EnumField(TranscriptionType)


class TranscriptionsSerializer(serializers.Serializer):
    """
    Allows for insertion of new transcriptions and zones
    in Bulk (uses serializer above) on a common parent
    """
    parent = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.filter(zone__isnull=False),
    )
    recognizer = MLToolField(MLToolType.Recognizer)
    transcriptions = TranscriptionBulkSerializer(many=True, allow_empty=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):  # May be None when generating an OpenAPI schema or using from a REPL
            return
        self.fields['parent'].queryset = Element.objects.filter(
            zone__isnull=False,
            corpus__in=Corpus.objects.writable(self.context['request'].user),
        )


class ClassificationBulkSerializer(serializers.Serializer):
    """
    Single classification serializer for bulk insertion
    Cannot use ModelSerializer as they become read-only when nested
    """
    class_name = serializers.CharField()
    confidence = serializers.FloatField(min_value=0, max_value=1)


class ClassificationsSerializer(serializers.Serializer):
    """
    Insert N classifications on a single element from a single ML tool
    """
    parent = serializers.PrimaryKeyRelatedField(
        queryset=Page.objects.filter(zone__isnull=False),
    )
    classifier = MLToolField(MLToolType.Classifier)
    classifications = ClassificationBulkSerializer(many=True, allow_empty=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):  # May be None when generating an OpenAPI schema or using from a REPL
            return
        self.fields['parent'].queryset = Page.objects.filter(
            zone__isnull=False,
            corpus__in=Corpus.objects.writable(self.context['request'].user),
        )
