from rest_framework import serializers
from arkindex.documents.models import Act, Page, TranscriptionType
from arkindex.documents.serializers.light import CorpusLightSerializer
from arkindex.documents.serializers.elements import ElementLightSerializer
from arkindex.documents.serializers.ml import TranscriptionSerializer
from arkindex.project.serializer_fields import EnumField, SimpleDateField, SearchTermsField


class SearchQuerySerializer(serializers.Serializer):
    """
    Search parameters validation serializer in order to build a ES query

    date_lte is rounded to superior value using ES syntax (see round_date()) and served as date_lt (exactly lower than)
    Rounded value depend if month/day are provided (Ex: '<= march 1999' becomes '< april 1999')
    """
    q = SearchTermsField(source='query')
    score = serializers.FloatField(source='min_score', min_value=0.0, max_value=1.0, default=0.0)
    date_gte = SimpleDateField(default='')
    date_lte = SimpleDateField(source='date_lt', rounded=True, default='')
    type = EnumField(enum=TranscriptionType, default='')
    corpus = serializers.UUIDField(source='corpus_id', default='')


class PageSearchResultSerializer(serializers.ModelSerializer):
    """
    A page search result with nested transcriptions
    """
    name = serializers.CharField(source='display_name')
    transcriptions = TranscriptionSerializer(many=True, source='transcriptions_results')
    total_transcriptions = serializers.IntegerField()
    parent_paths = serializers.ListField(
        child=serializers.ListField(
            child=ElementLightSerializer()
        ),
    )
    corpus = CorpusLightSerializer()

    class Meta:
        model = Page
        fields = (
            'id',
            'name',
            'transcriptions',
            'total_transcriptions',
            'parent_paths',
            'corpus',
        )


class ActSearchResultSerializer(serializers.ModelSerializer):
    """
    Serialize an act
    """
    transcriptions = TranscriptionSerializer(many=True, source='transcriptions_results')
    total_transcriptions = serializers.IntegerField()
    parent_paths = serializers.ListField(
        child=serializers.ListField(
            child=ElementLightSerializer()
        ),
    )
    corpus = CorpusLightSerializer()

    class Meta:
        model = Act
        fields = (
            'id',
            'name',
            'number',
            'transcriptions',
            'total_transcriptions',
            'parent_paths',
            'corpus',
        )
