from arkindex.project.polygon import Polygon
from arkindex.documents.models import Corpus, ElementType, Element, Page
from arkindex.documents.importer import parse_folio
from arkindex.images.models import Zone
from pathlib import Path
import xml.etree.ElementTree as ElementTree
import re
import logging

logger = logging.getLogger(__name__)

NS_PAGE = 'http://schema.primaresearch.org/PAGE/gts/pagecontent/2013-07-15'
NS = {'page': NS_PAGE}
NS_DELETE = "{%s}" % NS_PAGE

REGEX_VOLUME_NAME = re.compile(r'(?:JJ)?([0-9]+)([A-Z]?)')
REGEX_ACT_NUMBER = re.compile(r'([0-9]+)\s*([a-z]*)')
REGEX_FOLIO = re.compile(r'(?:fol\.?\s*)?([0-9]+)\s*(bis)?\s*(r|v)?', re.IGNORECASE)
REGEX_FOLIO_GROUP = re.compile(r'([^-_à]+)', re.IGNORECASE)


class SurfaceParser(object):
    def __init__(self, path):

        self.path = str(path)
        self.tree = ElementTree.parse(self.path)

        # Check we have a page
        self.page = self.tree.find('page:Page', NS)
        assert self.page is not None, 'Missing page'

    def get_metadata(self):

        # From metadata
        metadata = self.tree.find('page:Metadata', NS)
        assert metadata is not None, 'Missing metadata'
        dictMetadata = {node.tag.replace(NS_DELETE, "").lower(): node.text for node in metadata}

        # From page
        dictMetadata.update(self.page.attrib)

        return dictMetadata

    def list_surfaces(self):
        region_indexes = None

        # Load base elements from page
        read = self.page.find('page:ReadingOrder', NS)
        if read is not None:
            order = read.find('page:OrderedGroup', NS)
            assert order is not None, 'Missing ordered group'

            # Find indexes for each region
            region_indexes = {
                refindex.get('regionRef'): int(refindex.get('index'))
                for refindex in order.findall('page:RegionRefIndexed', NS)
            }

        out = []
        for i, textRegion in enumerate(self.page.findall('page:TextRegion', NS)):

            # Build coordinates polygon
            coords = textRegion.find('page:Coords', NS)
            assert coords is not None, 'Missing coords'
            polygon = Polygon([
                (int(x), int(y))
                for x, y in re.findall(r'(\d+),(\d+)', coords.get('points'))
            ])

            text_id = textRegion.get('id')
            if region_indexes is not None:  # If ReadingOrder exists
                # Find index for this text region
                index = region_indexes.get(text_id)
                assert index is not None, 'Missing index in region_indexes'
            else:
                index = i

            # Build output structure
            out.append({
                'index': index,
                'id': text_id,
                'polygon': polygon,
            })

        # Sort the surfaces by index
        return sorted(out, key=lambda x: x['index'])


class SurfaceImporter(object):

    def __init__(self, xmlpath=None, basepath=None, corpus=None, **kwargs):
        xmlpath = Path(xmlpath)
        assert isinstance(corpus, Corpus)
        assert xmlpath.exists()
        if basepath:
            basepath = Path(basepath)
            assert basepath.is_dir(), str(basepath)
            self.basepath = basepath
        else:
            self.basepath = xmlpath
        self.corpus = corpus
        self.xmlpath = xmlpath
        self.volumes = {}

    def get_xml_paths(self):
        """
        Get (volume name, path) tuples for XML files
        """
        if self.xmlpath.is_file():
            path = self.xmlpath.relative_to(self.basepath)
            yield path.parts[-2], path
            return

        for path in self.xmlpath.glob('**/*.xml'):
            path = path.relative_to(self.basepath)
            if len(path.parts) <= 1:  # File is in root folder
                logger.warning('File {} is not in a subfolder'.format(str(path)))
                continue
            yield path.parts[0], path

    def find_volume(self, name):
        """
        Get an Element instance from a volume name by case-insensitive match in the corpus
        """
        if name in self.volumes:
            return self.volumes[name]
        try:
            self.volumes[name] = self.corpus.elements.get(type=ElementType.Volume, name__iexact=name)
            logger.debug('Matched volume "{}" for volume name "{}"'.format(self.volumes[name], name))
            return self.volumes[name]
        except Element.DoesNotExist:
            logger.warning('No matching volume found for "{}"'.format(name))
        except Element.MultipleObjectsReturned:
            logger.warning('Multiple matching volumes found for "{}"'.format(name))

    def get_xml_pages(self):
        """
        Get (Page, path) tuples for XML files
        """
        for volume_name, path in self.get_xml_paths():
            volume = self.find_volume(volume_name)
            if not volume:
                continue
            page_type, nb, direction, complement = parse_folio(path.stem)
            try:
                page = Page.objects \
                    .get_descending(volume.id) \
                    .prefetch_related('zone__image') \
                    .get(
                        page_type=page_type,
                        nb=nb,
                        direction=direction,
                        complement=complement,
                    )
                yield page, path
            except ValueError:
                logger.warning("Could not parse folio '{}' in volume '{}'".format(path.stem, volume_name))
            except Page.DoesNotExist:
                logger.warning("No page found for folio '{}' in volume '{}'".format(path.stem, volume_name))
            except Page.MultipleObjectsReturned:
                logger.warning(
                    "Multiple matching pages found for folio '{}' in volume '{}'".format(path.stem, volume_name)
                )

    def run(self, dry=False):
        xml_count, surfaces_count, created_surfaces_count = 0, 0, 0

        for page, path in self.get_xml_pages():
            xml_count += 1
            s = SurfaceParser(self.basepath / path)
            image = page.zone.image

            if dry:  # Handle dry run
                logger.info("{}\t{}".format(path, image.path))
                continue

            for surfacedata in s.list_surfaces():
                poly = surfacedata['polygon']
                try:
                    z, _ = image.zones.get_or_create(polygon=poly)
                except ValueError as e:
                    logger.warning(
                        "Could not import zone with polygon '{}' on image '{}' from file '{}': {}".format(
                            poly, image.id, path, e,
                        )
                    )
                except Zone.MultipleObjectsReturned:
                    logger.warning('Multiple zones found, picking the first one')
                    z = image.zones.filter(polygon=poly).first()

                surface, created = Element.objects.get_or_create(
                    type=ElementType.Surface,
                    zone=z,
                    corpus=page.corpus,
                    defaults={'name': "Surface {}".format(surfacedata['id'])},
                )
                surface.add_parent(page)
                surfaces_count += 1
                created_surfaces_count += created

        logger.info("Parsed {} and created {} surfaces from {} XML files".format(
            surfaces_count, created_surfaces_count, xml_count,
        ))
