from unittest.mock import patch, ANY
from arkindex.documents.date_parser import DateParser, year, year_month, year_month_day, year_month_str
from arkindex.documents.models import DateType, InterpretedDate
from django.test import TestCase


class TestDateParser(TestCase):

    def _create_date(self, elts):
        return InterpretedDate(
            year=elts[0],
            month=elts[1],
            day=elts[2],
            type=elts[3],
        )

    @classmethod
    def setUpTestData(cls):
        cls.parser = DateParser()
        cls.test_dates = {
            '1343': ((1343, None, None, DateType.Exact), ),
            '1326-05': ((1326, 5, None, DateType.Exact), ),
            '1420-01-23': ((1420, 1, 23, DateType.Exact), ),
            '1398 february': ((1398, 2, None, DateType.Exact), ),
            'Août 1300': ((1300, 8, None, DateType.Exact), ),
            'march 12 11223344': (),
            '2222-22-22': (),
        }

    def test_parser(self):
        for t in self.test_dates:
            self.assertEqual(
                self.parser.parse(t),
                tuple([self._create_date(elts) for elts in self.test_dates[t]])
            )

    def _wrong_function(raw_date):
        return 'not a dict'

    @patch('logging.Logger.warning')
    def test_parser_exception(self, warning_mock):
        """
        Check that a error-raising function do not break table iteration
        """
        self.parser.date_table = (self._wrong_function, ) + self.parser.date_table
        self.assertEqual(self.parser.parse('1343')[0], InterpretedDate(year=1343, type=DateType.Exact))
        warning_mock.assert_called_with(ANY, self._wrong_function)

    def test_year(self):
        dates = ('1343', )
        wrong_dates = ('1343-03', )
        for d in dates:
            self.assertIsInstance(year(d), dict)
        for d in wrong_dates:
            self.assertIsNone(year(d))

    def test_year_month(self):
        dates = ('1343-02', )
        wrong_dates = ('1343-03-26', '1546')
        for d in dates:
            self.assertIsInstance(year_month(d), dict)
        for d in wrong_dates:
            self.assertIsNone(year_month(d))

    def test_year_month_day(self):
        dates = ('1343-2-5', )
        wrong_dates = ('1343-03', '1088')
        for d in dates:
            self.assertIsInstance(year_month_day(d), dict)
        for d in wrong_dates:
            self.assertIsNone(year_month_day(d))

    def test_year_month_str(self):
        dates = ('1343 july', '23 october 1245')
        wrong_dates = ('1343-03', 'july')
        for d in dates:
            self.assertIsInstance(year_month_str(d), dict)
        for d in wrong_dates:
            self.assertIsNone(year_month_str(d))
