from arkindex.project.tests import FixtureTestCase
from arkindex.documents.models import Element, ElementPath, ElementType
import itertools


class TestEditElementPath(FixtureTestCase):
    """
    Test ElementPath editing algorithms
    """

    def build_tree(self, tree, element_type=ElementType.Volume):
        """
        Build Element and ElementPath instances from a tree described by
        a dict associating elements with their direct ancestors.

        {'child': ['mother', 'father'], 'grandchild': 'child'}

        Returns a dict associating element names with created Elements.
        """
        assert isinstance(tree, dict)

        def parse_value(val):
            if isinstance(val, str):
                return [val]
            return val

        # List all elements
        # Pick all the keys, then add elements in the tree's values
        # Removes the need to add 'element': [] for root elements
        names = set(tree.keys()).union(*[parse_value(val) for val in tree.values()])
        # Create elements
        elements = {
            name: Element.objects.create(
                corpus=self.corpus,
                name=name,
                type=element_type,
            )
            for name in names
        }
        # Create paths
        for child_name, value in tree.items():
            parent_names = parse_value(value)
            for parent_name in parent_names:
                elements[child_name].add_parent(elements[parent_name])
        return elements

    def check_parents(self, elements, element_name, *paths):
        """
        Check an element's parent paths by names
        """
        tree = {
            elt_id: [e.path for e in elts]
            for elt_id, elts in itertools.groupby(
                ElementPath.objects.all().order_by('element_id'),
                lambda p: p.element_id,
            )
        }

        if paths is None or len(paths) < 1:
            self.assertNotIn(elements[element_name].id, tree)
            return

        return self.assertCountEqual(
            tree[elements[element_name].id],
            [
                [elements[parent_name].id for parent_name in path]
                for path in paths
            ]
        )

    def test_add_parent(self):
        r"""
        Test adding A as a parent of B.
            X   Y
             \ /
          K   A
           \ + \
            B   C
           / \   \
          D   E   F
         /
        G
        """
        elements = self.build_tree({
            'A': ['X', 'Y'],
            'C': 'A',
            'F': 'C',
            'B': 'K',
            'D': 'B',
            'E': 'B',
            'G': 'D',
        })

        self.check_parents(elements, 'X')
        self.check_parents(elements, 'Y')
        self.check_parents(elements, 'K')
        self.check_parents(elements, 'A', ['X'],
                                          ['Y'])
        self.check_parents(elements, 'B', ['K'])
        self.check_parents(elements, 'C', ['X', 'A'],
                                          ['Y', 'A'])
        self.check_parents(elements, 'D', ['K', 'B'])
        self.check_parents(elements, 'E', ['K', 'B'])
        self.check_parents(elements, 'F', ['X', 'A', 'C'],
                                          ['Y', 'A', 'C'])
        self.check_parents(elements, 'G', ['K', 'B', 'D'])

        elements['B'].add_parent(elements['A'])

        self.check_parents(elements, 'X')
        self.check_parents(elements, 'Y')
        self.check_parents(elements, 'K')
        self.check_parents(elements, 'A', ['X'],
                                          ['Y'])
        self.check_parents(elements, 'B', ['X', 'A'],
                                          ['Y', 'A'],
                                          ['K'])
        self.check_parents(elements, 'C', ['X', 'A'],
                                          ['Y', 'A'])
        self.check_parents(elements, 'D', ['X', 'A', 'B'],
                                          ['Y', 'A', 'B'],
                                          ['K', 'B'])
        self.check_parents(elements, 'E', ['X', 'A', 'B'],
                                          ['Y', 'A', 'B'],
                                          ['K', 'B'])
        self.check_parents(elements, 'F', ['X', 'A', 'C'],
                                          ['Y', 'A', 'C'])
        self.check_parents(elements, 'G', ['X', 'A', 'B', 'D'],
                                          ['Y', 'A', 'B', 'D'],
                                          ['K', 'B', 'D'])

    def test_add_parent_cycle(self):
        """
        Check we can't create cycles with add_parent
        A <-- B <-- C (parent <-- child)
        Tries to add C as a parent of A.
        """
        elements = self.build_tree({
            'C': 'B',
            'B': 'A',
        })
        with self.assertRaises(ValueError):
            elements['A'].add_parent(elements['C'])

    def test_add_parent_empty(self):
        """
        Check adding a parent works with an empty tree
        """
        elements = {
            'A': Element.objects.create(corpus=self.corpus, name="A", type=ElementType.Volume),
            'B': Element.objects.create(corpus=self.corpus, name="B", type=ElementType.Volume),
        }
        self.check_parents(elements, 'A')
        self.check_parents(elements, 'B')
        elements['B'].add_parent(elements['A'])
        self.check_parents(elements, 'A')
        self.check_parents(elements, 'B', ['A'])

    def test_remove_child(self):
        r"""
        Test removing child B from parent A.
            X   Y
             \ /
          K   A
           \ - \
            B   C
           / \   \
          D   E   F
         /
        G
        """
        elements = self.build_tree({
            'A': ['X', 'Y'],
            'C': 'A',
            'F': 'C',
            'B': ['A', 'K'],
            'D': 'B',
            'E': 'B',
            'G': 'D',
        })

        self.check_parents(elements, 'X')
        self.check_parents(elements, 'Y')
        self.check_parents(elements, 'K')
        self.check_parents(elements, 'A', ['X'],
                                          ['Y'])
        self.check_parents(elements, 'B', ['X', 'A'],
                                          ['Y', 'A'],
                                          ['K'])
        self.check_parents(elements, 'C', ['X', 'A'],
                                          ['Y', 'A'])
        self.check_parents(elements, 'D', ['X', 'A', 'B'],
                                          ['Y', 'A', 'B'],
                                          ['K', 'B'])
        self.check_parents(elements, 'E', ['X', 'A', 'B'],
                                          ['Y', 'A', 'B'],
                                          ['K', 'B'])
        self.check_parents(elements, 'F', ['X', 'A', 'C'],
                                          ['Y', 'A', 'C'])
        self.check_parents(elements, 'G', ['X', 'A', 'B', 'D'],
                                          ['Y', 'A', 'B', 'D'],
                                          ['K', 'B', 'D'])

        elements['A'].remove_child(elements['B'])

        self.check_parents(elements, 'X')
        self.check_parents(elements, 'Y')
        self.check_parents(elements, 'K')
        self.check_parents(elements, 'A', ['X'],
                                          ['Y'])
        self.check_parents(elements, 'B', ['K'])
        self.check_parents(elements, 'C', ['X', 'A'],
                                          ['Y', 'A'])
        self.check_parents(elements, 'D', ['K', 'B'])
        self.check_parents(elements, 'E', ['K', 'B'])
        self.check_parents(elements, 'F', ['X', 'A', 'C'],
                                          ['Y', 'A', 'C'])
        self.check_parents(elements, 'G', ['K', 'B', 'D'])

    def test_remove_child_update(self):
        r"""
        Scenario where updates are required
            X   Y
             \ /
              A
             - \
            B   C
           / \
          D   E
               \
                F
        """

        elements = self.build_tree({
            'A': ['X', 'Y'],
            'B': 'A',
            'C': 'A',
            'D': 'B',
            'E': 'B',
            'F': 'E',
        })

        self.check_parents(elements, 'X')
        self.check_parents(elements, 'Y')
        self.check_parents(elements, 'A', ['X'],
                                          ['Y'])
        self.check_parents(elements, 'B', ['X', 'A'],
                                          ['Y', 'A'])
        self.check_parents(elements, 'C', ['X', 'A'],
                                          ['Y', 'A'])
        self.check_parents(elements, 'D', ['X', 'A', 'B'],
                                          ['Y', 'A', 'B'])
        self.check_parents(elements, 'E', ['X', 'A', 'B'],
                                          ['Y', 'A', 'B'])
        self.check_parents(elements, 'F', ['X', 'A', 'B', 'E'],
                                          ['Y', 'A', 'B', 'E'])

        elements['A'].remove_child(elements['B'])

        self.check_parents(elements, 'X')
        self.check_parents(elements, 'Y')
        self.check_parents(elements, 'B')
        self.check_parents(elements, 'A', ['X'],
                                          ['Y'])
        self.check_parents(elements, 'C', ['X', 'A'],
                                          ['Y', 'A'])
        self.check_parents(elements, 'D', ['B'])
        self.check_parents(elements, 'E', ['B'])
        self.check_parents(elements, 'F', ['B', 'E'])

    def test_delete_element(self):
        r"""
        Test the signal works properly and is able to remove all paths when removing an Element.
            X   Y
             \ /
              A
             / \
           [B]  C
           / \
          D   E
               \
                F
        """

        elements = self.build_tree({
            'A': ['X', 'Y'],
            'B': 'A',
            'C': 'A',
            'D': 'B',
            'E': 'B',
            'F': 'E',
        })

        self.check_parents(elements, 'X')
        self.check_parents(elements, 'Y')
        self.check_parents(elements, 'A', ['X'],
                                          ['Y'])
        self.check_parents(elements, 'B', ['X', 'A'],
                                          ['Y', 'A'])
        self.check_parents(elements, 'C', ['X', 'A'],
                                          ['Y', 'A'])
        self.check_parents(elements, 'D', ['X', 'A', 'B'],
                                          ['Y', 'A', 'B'])
        self.check_parents(elements, 'E', ['X', 'A', 'B'],
                                          ['Y', 'A', 'B'])
        self.check_parents(elements, 'F', ['X', 'A', 'B', 'E'],
                                          ['Y', 'A', 'B', 'E'])

        elements['B'].delete()

        self.check_parents(elements, 'X')
        self.check_parents(elements, 'Y')
        self.check_parents(elements, 'A', ['X'],
                                          ['Y'])
        self.check_parents(elements, 'C', ['X', 'A'],
                                          ['Y', 'A'])
        self.check_parents(elements, 'D')
        self.check_parents(elements, 'E')
        self.check_parents(elements, 'F', ['E'])
