from django.test import TestCase
from unittest.mock import patch, call, MagicMock
from elasticsearch import Elasticsearch
from elasticsearch.exceptions import NotFoundError
from arkindex.documents.indexer import Indexer


class TestIndexer(TestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.fake_items = [MagicMock() for _ in range(10)]
        for item in cls.fake_items:
            item.id.hex = 'id'
            item.build_search_index.return_value = {'text': 'hello'}

    @patch('arkindex.documents.indexer.Elasticsearch')
    def test_drop_index(self, es_mock):
        indexer = Indexer(['a', 'b'])
        self.assertEqual(es_mock.call_count, 1)
        self.assertEqual(es_mock.call_args, call(['a', 'b']))
        indexer.drop_index('something')
        self.assertEqual(es_mock().indices.delete.call_count, 1)
        self.assertEqual(es_mock().indices.delete.call_args, call(index='something'))

    @patch('arkindex.documents.indexer.Elasticsearch')
    def test_drop_index_handle_exceptions(self, es_mock):
        # Catches NotFoundError
        es_mock.return_value.indices.delete.side_effect = NotFoundError
        Indexer().drop_index('something')
        self.assertEqual(es_mock.call_count, 1)
        self.assertEqual(es_mock().indices.delete.call_count, 1)
        self.assertEqual(es_mock().indices.delete.call_args, call(index='something'))

        # Raises anything else
        es_mock.reset_mock()
        es_mock.return_value.indices.delete.side_effect = ValueError
        with self.assertRaises(ValueError):
            Indexer().drop_index('something')
        self.assertEqual(es_mock.call_count, 1)
        self.assertEqual(es_mock().indices.delete.call_count, 1)
        self.assertEqual(es_mock().indices.delete.call_args, call(index='something'))

    @patch('arkindex.documents.indexer.es_bulk')
    def test_run_index(self, es_bulk_mock):
        es_bulk_mock.return_value = (10, [])
        queryset_mock = MagicMock()
        queryset_mock.__getitem__().all.return_value = self.fake_items
        queryset_mock.count.return_value = len(self.fake_items) * 10

        Indexer().run_index('name', 'type', queryset_mock, bulk_size=10)

        self.assertEqual(es_bulk_mock.call_count, 10)
        for (es_instance, actions), kwargs in es_bulk_mock.call_args_list:
            self.assertIsInstance(es_instance, Elasticsearch)
            self.assertListEqual(
                actions,
                [{
                    '_index': 'name',
                    '_type': 'type',
                    '_source': {'text': 'hello'},
                    '_id': 'id',
                }] * 10
            )
            self.assertDictEqual(kwargs, {
                'chunk_size': 100,
                'initial_backoff': 2,
                'max_retries': 2,
                'request_timeout': 15,
                'stats_only': True,
                '_source': False,
            })

    @patch('arkindex.documents.indexer.es_bulk')
    def test_index_retry(self, es_bulk_mock):
        queryset_mock = MagicMock()
        queryset_mock.all.return_value = self.fake_items
        queryset_mock.count.return_value = len(self.fake_items)
        # Keep failing
        es_bulk_mock.side_effect = Exception

        count = Indexer().index('name', 'type', queryset_mock, chunk_size=10, timeout=42)

        self.assertEqual(count, 0)
        self.assertEqual(es_bulk_mock.call_count, 3)
        args, kwargs = zip(*es_bulk_mock.call_args_list)

        for es_instance, actions in args:
            self.assertIsInstance(es_instance, Elasticsearch)
            self.assertListEqual(
                actions,
                [{
                    '_index': 'name',
                    '_type': 'type',
                    '_source': {'text': 'hello'},
                    '_id': 'id',
                }] * 10
            )

        self.assertTupleEqual(kwargs, (
            {
                'chunk_size': 10,
                'initial_backoff': 2,
                'max_retries': 2,
                'request_timeout': 42,
                'stats_only': True,
                '_source': False,
            },
            {
                'chunk_size': 5,
                'initial_backoff': 2,
                'max_retries': 2,
                'request_timeout': 84,
                'stats_only': True,
                '_source': False,
            },
            {
                'chunk_size': 2,
                'initial_backoff': 2,
                'max_retries': 2,
                'request_timeout': 168,
                'stats_only': True,
                '_source': False,
            },
        ))
