from lxml import etree
from arkindex.documents.models import Act
from arkindex.documents.tei import Text, TeiParser
from arkindex.dataimport.models import DataImportFailure
from arkindex.project.tests import FixtureTestCase
import os.path

FIXTURES = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    'tei_samples',
)


class TestTeiElement(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.repo = cls.user.credentials.get().repos.get()

    def test_apply_xslt(self):
        tree = etree.parse(os.path.join(FIXTURES, 'arguments.xml'))
        te = Text(tree.getroot())
        self.assertMultiLineEqual(
            # Remove indentations
            "".join(line.strip() for line in te.argument.splitlines()),
            '<div xmlns:tei="http://www.tei-c.org/ns/1.0" class="argument">'
            '<p>Argument</p>'
            '<p><em>Italique</em></p>'
            '<p><sup>Supérieur</sup></p>'
            '<p><sup>Supérieur</sup></p>'
            '<p><blockquote>Citation</blockquote></p>'
            '</div>')

    def test_save_update(self):
        """
        Check saving metadata for a TeiElement updates existing data
        """
        self.maxDiff = None
        act = Act.objects.get(number="1")
        tree_before = etree.parse(os.path.join(FIXTURES, 'update_before.xml'))
        tree_after = etree.parse(os.path.join(FIXTURES, 'update_after.xml'))
        te_before = Text(tree_before.getroot())
        te_after = Text(tree_after.getroot())

        rev1 = self.repo.revisions.get()
        rev2 = self.repo.revisions.create(hash='43', ref='ref/heads/master', message='b', author='me')

        te_before.save(act, rev1)
        self.assertFalse(act.metadatas.exclude(revision=rev1).exists())
        self.assertCountEqual(
            act.metadatas.exclude(name='summary').values_list('name', 'index', 'value'),
            [
                ('folio', 0, '1r'),
                ('date', 0, '1337-01'),
                ('id', 0, 'test_update'),
                ('language', 0, 'lat.'),
                ('name', 0, '1'),
                ('location', 0, 'Mars'),
                ('person', 0, 'Someone'),
                ('person', 1, 'The Chosen One'),
                ('place', 0, 'Somewhere'),
            ]
        )

        # Create subject, update person and delete place, leave location untouched
        te_after.save(act, rev2)
        self.assertCountEqual(
            act.metadatas.exclude(name='summary').values_list('name', 'index', 'value', 'revision_id'),
            [
                ('folio', 0, '1r', rev1.id),
                ('date', 0, '1337-01', rev1.id),
                ('id', 0, 'test_update', rev1.id),
                ('language', 0, 'lat.', rev1.id),
                ('name', 0, '1', rev1.id),
                ('location', 0, 'Mars', rev1.id),
                ('person', 0, 'Someone else', rev2.id),
                ('person', 1, 'The Chosen One', rev1.id),
                ('subject', 0, 'Something', rev2.id),
            ]
        )

    def test_match_error(self):
        xml_path = os.path.join(FIXTURES, 'arguments.xml')
        parser = TeiParser(xml_path)
        parser._match_error(parser.corpus, 'Some funny test message')
        self.assertEqual(len(parser.match_errors), 1)
        failure = parser.match_errors[0]
        self.assertIsInstance(failure, DataImportFailure)
        self.assertEqual(failure.path, xml_path)
        self.assertEqual(failure.line, 1)
        self.assertEqual(failure.message, 'Some funny test message')
        self.assertIsNone(failure.dataimport_id)
        self.assertIsNone(failure.element_id)
