from django.db import models
from django.db.models.functions import Concat, Substr
from django.conf import settings
from django.utils.functional import cached_property
from django.utils.text import slugify
from arkindex.documents.models import Page
from arkindex.images.managers import ImageServerManager
from arkindex.project.models import IndexableModel
from arkindex.project.fields import StripSlashURLField, LStripTextField
from arkindex.project.polygon import PolygonField
from arkindex.project.aws import s3
from botocore.client import ClientError
from enumfields import EnumField, Enum
from io import BytesIO
from PIL import Image as PillowImage
import logging
import os
import requests
import urllib.parse

logger = logging.getLogger(__name__)


class ImageServer(models.Model):
    """
    A IIIF image server
    """
    display_name = models.CharField(max_length=250)
    url = StripSlashURLField(unique=True)

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    validated = models.BooleanField(default=False)

    objects = ImageServerManager()

    def __str__(self):
        return self.display_name

    @property
    def is_local(self):
        """
        Indicates whether the server is the Arkindex IIIF server.
        """
        return self.id == settings.LOCAL_IMAGESERVER_ID

    def find_image(self, path, offline=False, width=None, height=None, save=True):
        """
        Lookup an image on a server
        This is the preferred way to construct an image
        """
        img = None
        # Try direct access to path
        img = self.images.filter(path=path).first()

        if img is None:
            # Try the url encoded path
            img = self.images.filter(
                path=urllib.parse.quote_plus(path),
            ).first()

        if img is None:
            img = Image(server=self, path=path, width=width, height=height)

        # Support offline queries
        if not offline:
            img.perform_check(save=save)

        if save:
            img.save()

        return img

    def build_url(self, path):
        """
        Build a full url on this server
        """
        base_url = self.url
        path = path.lstrip('/')
        if not base_url.endswith('/'):
            base_url += '/'
        abs_url = urllib.parse.urljoin(base_url, path)

        # In some cases, the path will contain something that looks like a scheme,
        # e.g. ark:/something with Gallica URLs. In those cases, urljoin will not add
        # the server's root URL ; we fall back to a dirty way of joining the URLs.
        if abs_url.startswith(base_url):
            return abs_url

        return base_url + path  # Don't try this at home

    def merge_into(self, other):
        """
        Merge this ImageServer into another. This does not remove the ImageServer,
        it only moves around the images.
        This allows changing the root folder of a server and updating all of its images.
        Both servers must have the same URL schemes and domains.
        If both servers have the same root path, the images are simply moved.
        If the destination server is a parent folder of this server, the subfolder is prepended to all image paths.
        If the destination server is a child folder of this server, the subfolder is removed from all paths.
        In the latter case, merging is only allowed when all images of the server are in the same subfolder.
        Raises ValueError whenever merging is impossible.
        """
        assert isinstance(other, ImageServer)
        if not self.images.exists():
            return

        # Parse both URLs to first check for scheme/netloc compatibility
        src_url, dest_url = map(urllib.parse.urlsplit, (self.url, other.url))
        if src_url.scheme != dest_url.scheme:
            raise ValueError('Cannot merge into an image server of a different protocol')
        elif src_url.netloc != dest_url.netloc:
            raise ValueError('Cannot merge into an image server on a different domain')

        # Check paths
        src_path, dest_path = src_url.path.strip('/'), dest_url.path.strip('/')
        common = os.path.commonpath([src_path, dest_path])

        if common == dest_path:
            # Destination is a parent folder of this server
            prefix = src_path[len(common):].lstrip('/')
            if not prefix.endswith('/'):
                prefix += '/'
            # Prepend all image paths with this new prefix and change their server
            self.images.update(
                server_id=other.id,
                path=Concat(models.Value(prefix), 'path'),
            )
            return

        if common == src_path:
            # Destination is a child of this server
            prefix = dest_path[len(common):].lstrip('/')
            if not prefix.endswith('/'):
                prefix += '/'
            # Check that all images on this server start with the prefix
            if self.images.exclude(path__startswith=prefix).exists():
                raise ValueError('Cannot merge into an image server on a child folder: '
                                 'Some images are not in the child folder')
            # Remove the prefix from all image paths and change their server
            # Note that positions in Substr() are 1-indexed
            self.images.update(
                server_id=other.id,
                path=Substr('path', len(prefix) + 1),
            )
            return

        raise ValueError('Cannot merge into an image server on a different subfolder')

    def split(self, folder):
        """
        Create a new ImageServer for a subfolder on this server and move the images there.
        Raises ValueError if there are no images to move.
        Returns the newly created ImageServer.
        """
        # Ensure we have a path ending with a single slash
        folder_slash = folder.lstrip('/') + '/'
        images = self.images.filter(path__startswith=folder_slash)
        if not images.exists():
            raise ValueError('No images are in the "{}" folder'.format(folder))

        # Create the new server on the subfolder
        new_server = ImageServer.objects.create(
            display_name='{}_{}'.format(self.display_name, slugify(folder)),
            url=urllib.parse.urljoin(self.url.lstrip('/') + '/', folder),
        )

        # Move the images there, removing the subfolder from their path
        images.update(
            server_id=new_server.id,
            # Substr positions are 1-indexed
            path=Substr('path', len(folder_slash) + 1)
        )
        return new_server

    @property
    def grafana_tag(self):
        return '{}_{}'.format(self.display_name, self.id)


class ImageStatus(Enum):
    """
    Image validation status
    """
    Checked = "checked"
    Unchecked = "unchecked"
    Error = "error"


class Image(IndexableModel):
    """
    A document image
    """
    server = models.ForeignKey(ImageServer, on_delete=models.CASCADE, related_name='images')
    path = LStripTextField(chars='/')
    width = models.PositiveIntegerField(default=0)
    height = models.PositiveIntegerField(default=0)
    status = EnumField(ImageStatus, default=ImageStatus.Unchecked, max_length=50)

    class Meta:
        unique_together = (
            ('server', 'path'),
        )

    @property
    def url(self):
        return self.server.build_url(self.path)

    @cached_property
    def s3_object(self):
        assert self.server.is_local, 'Cannot load images on remote image servers via S3'
        return s3.Object(settings.AWS_IIIF_BUCKET, urllib.parse.unquote(self.path))

    @property
    def s3_url(self):
        return s3.meta.client.generate_presigned_url('get_object', Params={
            'Bucket': self.s3_object.bucket_name,
            'Key': self.s3_object.key,
        })

    def exists(self):
        """
        Returns whether the Image exists on the IIIF S3 bucket by performing a HEAD request to S3.
        See https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Object.load
        """
        try:
            self.s3_object.load()
            return True
        except ClientError as e:
            if e.response['Error']['Code'] != '404':
                raise
            return False

    def get_thumbnail_url(self, max_width=200, max_height=None):
        if max_width is None and max_height is None:
            resize = "full"
        else:
            resize = "{0},{1}".format(max_width or '', max_height or '')
        return "{0}/full/{1}/0/default.jpg".format(self.url.rstrip('/'), resize)

    def pillow_open(self, max_width=500):
        if self.server.is_local:
            b = BytesIO()
            self.s3_object.download_fileobj(b)
            return PillowImage.open(b)

        # PIL.Image.open explicitly requires seek(int) method, that the urllib responses do not provide.
        # We therefore have to get the whole content and put it back in a file-like object
        resp = requests.get(
            self.get_thumbnail_url(max_width=max_width),
            timeout=settings.IIIF_DOWNLOAD_TIMEOUT,
        )
        resp.raise_for_status()
        return PillowImage.open(BytesIO(resp.content))

    def perform_check(self, save=True):
        """
        Check the image's existence and update width, height and status properties
        """

        url = self.url
        if not url.endswith('/'):
            url += '/'
        info_url = urllib.parse.urljoin(url, 'info.json')
        try:
            # Load info
            resp = requests.get(info_url, timeout=15, allow_redirects=True)
            resp.raise_for_status()

            # Load data
            data = resp.json()
            assert all(item in data for item in ('@id', 'width', 'height'))

            # Check id
            image_id = data['@id']
            assert image_id.startswith(self.server.url), \
                'Image id does not start with server url ({} vs. {})'.format(image_id, self.server.url)
        except Exception as e:
            logger.warn('Image check failed: {}'.format(str(e)))
            self.status = ImageStatus.Error
            if save:
                self.save()
            return

        # Use image resource ID from IIIF server to update the image path if needed
        self.path = image_id[len(self.server.url) + 1:]
        self.width, self.height = int(data['width']), int(data['height'])

        self.status = ImageStatus.Checked
        if save:
            self.save()

    def pillow_save(self, pillow_img, format=None):
        """
        Save a PIL.Image instance to S3
        """
        assert self.server.is_local, 'Only images on the local server can be saved'
        logger.debug('Saving {} to S3'.format(self.id))
        b = BytesIO()
        pillow_img.save(b, format=format)
        b.seek(0)
        self.s3_object.upload_fileobj(b)

    def __str__(self):
        return '{} - {}'.format(self.id, self.url)


class Thumbnail(object):
    """
    Describes an element thumbnail
    """

    def __init__(self, element):
        self.element = element

    @property
    def name(self):
        return '{}.jpg'.format(str(self.element.id.hex))

    @cached_property
    def s3_object(self):
        return s3.Object(settings.AWS_THUMBNAIL_BUCKET, self.name)

    @property
    def url(self):
        return s3.meta.client.generate_presigned_url('get_object', Params={
            'Bucket': self.s3_object.bucket_name,
            'Key': self.s3_object.key,
        })

    def create(self, width=900, height=400, max_element_count=3, force=False):
        """
        Generate a thumbnail for an Element and store it in the IIIF server
        """

        if not force:
            try:
                self.s3_object.load()
                # This did not raise anything so the thumbnail exists
                return
            except ClientError as e:
                if e.response['Error']['Code'] != '404':
                    raise

        # Get at most 'max_element_count' first pages
        pages = Page.objects \
            .get_descending(self.element.id) \
            .prefetch_related('zone__image__server')[:max_element_count]
        if not pages:
            raise Page.DoesNotExist("No pages found for thumbnail generation")

        # Width of a single image in the thumbnail
        single_width = int(width / len(pages))

        # Open images with Pillow
        images = [p.zone.image.pillow_open(max_width=single_width) for p in pages]

        # Create an image
        thumbnail = PillowImage.new('RGB', (width, height))

        # Resize, crop and assemble the images together horizontally
        offset = 0
        for img in images:
            # Resize
            imgwidth, imgheight = img.size
            ratio = max(single_width / imgwidth, height / imgheight)
            newsize = int(imgwidth * ratio), int(imgheight * ratio)
            img = img.resize(newsize, PillowImage.BICUBIC)

            # Crop
            imgwidth, imgheight = img.size
            left = int((imgwidth - single_width) / 2)
            top = int((imgheight - height) / 2)
            img = img.crop((left, top, left + single_width, top + height))

            # Assemble
            thumbnail.paste(img, (offset, 0))
            offset += single_width

        b = BytesIO()
        thumbnail.save(b, format='jpeg')
        b.seek(0)
        self.s3_object.upload_fileobj(b)


class Zone(IndexableModel):
    """
    A zone on an image
    """
    image = models.ForeignKey(Image, on_delete=models.CASCADE, related_name='zones')

    polygon = PolygonField()

    def __str__(self):
        return 'Zone {}'.format(self.polygon)

    @property
    def url(self):
        '''
        Build the IIIF crop url
        '''
        return urllib.parse.urljoin(
            self.image.url + '/',
            '{},{},{},{}/full/0/default.jpg'.format(
                self.polygon.x, self.polygon.y, self.polygon.width, self.polygon.height),
        )
