from arkindex.project.tests import FixtureTestCase
from arkindex.documents.models import ElementType, Page
from unittest.mock import patch, call
from django.test import override_settings
from botocore.exceptions import ClientError


@override_settings(LOCAL_IMAGESERVER_ID=999, IIIF_DOWNLOAD_TIMEOUT=(13, 37))
class TestThumbnail(FixtureTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.s3_mock = patch('arkindex.images.models.s3').start()

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.vol1 = cls.corpus.elements.get(name='Volume 1', type=ElementType.Volume)

    def tearDown(self):
        super().tearDown()
        self.s3_mock.reset_mock()
        try:
            # Remove the cached S3 object to force a reinstanciation
            del self.vol1.thumbnail.s3_object
        except AttributeError:  # Sometimes, it isn't already cached
            pass

    def test_name(self):
        self.assertEqual(self.vol1.thumbnail.name, str(self.vol1.id.hex) + '.jpg')

    def test_s3_object(self):
        with self.settings(AWS_THUMBNAIL_BUCKET="derp"):
            # Test twice to ensure s3_object is only created once
            self.assertEqual(self.vol1.thumbnail.s3_object, self.s3_mock.Object.return_value)
            self.assertEqual(self.vol1.thumbnail.s3_object, self.s3_mock.Object.return_value)
        self.assertEqual(self.s3_mock.Object.call_count, 1)
        self.assertEqual(self.s3_mock.Object.call_args, call('derp', self.vol1.thumbnail.name))

    def test_url(self):
        self.s3_mock.meta.client.generate_presigned_url.return_value = 'http://nowhere'
        self.s3_mock.Object.return_value.bucket_name = 'derp'
        self.s3_mock.Object.return_value.key = 'meme.jpg'
        with self.settings(AWS_THUMBNAIL_BUCKET="derp"):
            self.assertEqual(self.vol1.thumbnail.url, 'http://nowhere')
        self.assertEqual(self.s3_mock.Object.call_count, 1)
        self.assertEqual(self.s3_mock.meta.client.generate_presigned_url.call_count, 1)
        self.assertEqual(
            self.s3_mock.meta.client.generate_presigned_url.call_args,
            call('get_object', Params={'Bucket': 'derp', 'Key': 'meme.jpg'}),
        )

    @patch('arkindex.images.models.BytesIO')
    @patch('arkindex.images.models.PillowImage')
    @patch('arkindex.images.models.requests')
    def test_create(self, requests_mock, pil_mock, bytes_mock):
        """
        Test Thumbnail.create creates a thumbnail if it does not exist
        """
        self.s3_mock.Object.return_value.load.side_effect = ClientError({'Error': {'Code': '404'}}, 'get_object')
        requests_mock.get.return_value.content = b'something'
        pil_mock.open.return_value.size = (1000, 1000)
        pil_mock.open.return_value.resize.return_value.size = (400, 400)

        self.vol1.thumbnail.create()

        self.assertEqual(pil_mock.new.call_count, 1)
        self.assertEqual(pil_mock.new.call_args, call('RGB', (900, 400)))
        self.assertEqual(pil_mock.open.call_count, 3)
        self.assertEqual(requests_mock.get.call_count, 3)
        self.assertEqual(requests_mock.get.call_args_list, [
            call('http://server/img1/full/300,/0/default.jpg', timeout=(13, 37)),
            call('http://server/img2/full/300,/0/default.jpg', timeout=(13, 37)),
            call('http://server/img3/full/300,/0/default.jpg', timeout=(13, 37)),
        ])
        self.assertEqual(pil_mock.open().resize.call_count, 3)
        self.assertListEqual(
            pil_mock.open().resize.call_args_list,
            [call((400, 400), pil_mock.BICUBIC)] * 3,
        )
        self.assertEqual(pil_mock.open().resize().crop.call_count, 3)
        self.assertEqual(
            pil_mock.open().resize().crop.call_args_list,
            [call((50, 0, 350, 400))] * 3,
        )
        self.assertEqual(pil_mock.new().paste.call_count, 3)
        self.assertEqual(pil_mock.new().paste.call_args_list, [
            call(pil_mock.open().resize().crop(), (0, 0)),
            call(pil_mock.open().resize().crop(), (300, 0)),
            call(pil_mock.open().resize().crop(), (600, 0)),
        ])

        # BytesIO is used three times for opening images, then once for the thumbnail save
        self.assertEqual(bytes_mock.call_count, 4)
        self.assertEqual(bytes_mock.call_args_list, [call(b'something')] * 3 + [call()])
        self.assertEqual(pil_mock.new().save.call_count, 1)
        self.assertEqual(
            pil_mock.new().save.call_args,
            call(bytes_mock(), format='jpeg'),
        )
        self.assertEqual(bytes_mock().seek.call_count, 1)
        self.assertEqual(bytes_mock().seek.call_args, call(0))
        self.assertEqual(self.s3_mock.Object.call_count, 1)
        self.assertEqual(self.s3_mock.Object().upload_fileobj.call_count, 1)
        self.assertEqual(self.s3_mock.Object().upload_fileobj.call_args, call(bytes_mock()))

    @patch('arkindex.images.models.PillowImage')
    def test_create_exists(self, pil_mock):
        """
        Test Thumbnail.create ignores creation if the thumbnail exists (if S3 does not raise 404)
        """
        self.s3_mock.Object.return_value.load.side_effect = None
        self.vol1.thumbnail.create()
        self.assertEqual(self.s3_mock.Object.call_count, 1)
        self.assertEqual(self.s3_mock.Object().load.call_count, 1)
        self.assertEqual(self.s3_mock.Object().upload_fileobj.call_count, 0)
        self.assertEqual(pil_mock.new.call_count, 0)
        self.assertEqual(pil_mock.open.call_count, 0)

    def test_create_exception(self):
        """
        Test Thumbnail.create raises any error that is not 404
        """
        self.s3_mock.Object.return_value.load.side_effect = ClientError({'Error': {'Code': '999'}}, 'get_object')
        with self.assertRaises(ClientError):
            self.vol1.thumbnail.create()

    @patch('arkindex.images.models.BytesIO')
    @patch('arkindex.images.models.PillowImage')
    @patch('arkindex.images.models.requests')
    def test_create_force(self, requests_mock, pil_mock, bytes_mock):
        """
        Test Thumbnail.create ignores existing thumbnails with force=True
        """
        requests_mock.get.return_value.content = b'something'
        pil_mock.open.return_value.size = (1000, 1000)
        pil_mock.open.return_value.resize.return_value.size = (400, 400)

        self.vol1.thumbnail.create(force=True)

        self.assertEqual(self.s3_mock.Object.call_count, 1)
        self.assertEqual(self.s3_mock.Object().load.call_count, 0)
        self.assertEqual(pil_mock.new.call_count, 1)
        self.assertEqual(pil_mock.open.call_count, 3)
        self.assertEqual(self.s3_mock.Object().upload_fileobj.call_count, 1)
        self.assertEqual(self.s3_mock.Object().upload_fileobj.call_args, call(bytes_mock()))

    def test_create_empty(self):
        """
        Test Thumbnail.create fails if there are no pages
        """
        self.s3_mock.Object.return_value.load.side_effect = ClientError({'Error': {'Code': '404'}}, 'get_object')
        Page.objects.get_descending(self.vol1.id).delete()
        with self.assertRaises(Page.DoesNotExist):
            self.vol1.thumbnail.create()
