from django.conf import settings
from rest_framework import serializers
from enum import Enum
from arkindex_common.ml_tool import MLTool, MLToolType
from arkindex.project.tools import elasticsearch_escape
import calendar
import re

DATE_REGEX = re.compile(r'^(?P<year>[1-2]\d{3})(-(?P<month>0[1-9]|1[0-2])(-(?P<day>\d{2}))?)?$')


class EnumField(serializers.ChoiceField):
    """
    Serializes an enum field into a JSON value
    """

    def __init__(self, enum, *args, **kwargs):
        assert issubclass(enum, Enum) or not enum
        self.enum = enum
        choices = [(item.value, item.name) for item in self.enum]
        super().__init__(choices, *args, **kwargs)

    def to_representation(self, obj):
        return obj.value

    def to_internal_value(self, data):
        assert self.enum is not None, "No enum set on EnumField"
        try:
            return self.enum(data)
        except ValueError:
            raise serializers.ValidationError('Value is not of type {}'.format(self.enum.__name__))


class MLToolField(serializers.CharField):

    def __init__(self, tool_type, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert isinstance(tool_type, MLToolType)
        self.tool_type = tool_type

    def to_representation(self, obj):
        return obj.slug

    def to_internal_value(self, data):
        try:
            return MLTool.get(settings.ML_CLASSIFIERS_DIR, self.tool_type, data)
        except ValueError as e:
            raise serializers.ValidationError(str(e))


class SearchTermsField(serializers.CharField):
    """
    Serialize ElastiSearch query terms
    """
    def to_internal_value(self, query_terms):
        return elasticsearch_escape(query_terms)


def date_validator(date):
    """
    Validate that date format is understandable for ElasticSearch
    Regex will match a mandatory year then optional 2-digit month and day.
    Year (between 1000 and 2999) and month (between 1 and 12) are directly validated by the
    regex whereas day, if found, is directly taken as a 2-digit number and checked later.
    """
    match = DATE_REGEX.match(date)
    if not match:
        raise serializers.ValidationError('Could not parse date. format should be YYYY[-MM[-DD]]')
    date_dict = match.groupdict()
    # Validate day depending on year/month if a day is present
    if date_dict['day']:
        date_dict = {k: int(e) for k, e in date_dict.items()}
        if not date_dict['day'] in range(1, calendar.monthrange(date_dict['year'], date_dict['month'])[1] + 1):
            raise serializers.ValidationError('Could not parse date. Day is invalid')


def round_date(date):
    """
    Add ElasticSearch suffix to round date to superior depending if month/day are provided
    https://www.elastic.co/guide/en/elasticsearch/reference/6.2/common-options.html#date-math
    """
    precision = '||+1d'
    if re.match(r'^\d{4}-\d{2}$', date):
        precision = '||+1M'
    elif re.match(r'^\d{4}$', date):
        precision = '||+1y'
    return date + precision


class SimpleDateField(serializers.CharField):
    """
    Serialize a date understandable by ElasticSearch
    """
    def __init__(self, rounded=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rounded = rounded

    def to_internal_value(self, date):
        date_validator(date)
        if self.rounded:
            date = round_date(date)
        return date
