"""
Django settings for arkindex project.

Generated by 'django-admin startproject' using Django 1.11.6.

For more information on this file, see
https://docs.djangoproject.com/en/1.11/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.11/ref/settings/
"""

import os
import sys
import logging
from corsheaders.defaults import default_headers


def env2list(env_name, separator=',', default=[]):
    '''
    Load env variable as a list
    '''
    value = os.environ.get(env_name)
    return value and value.split(separator) or default


# Database
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'NAME': os.environ.get('DB_NAME', 'arkindex_dev'),
        'USER': os.environ.get('DB_USER', 'devuser'),
        'PASSWORD': os.environ.get('DB_PASSWORD', 'devdata'),
        'HOST': os.environ.get('DB_HOST', 'localhost'),
        'PORT': os.environ.get('DB_PORT', 9100),
    }
}

# Admins in charge
ADMINS = [
    ('Bastien', 'abadie@teklia.com'),
    ('Erwan', 'rouchet@teklia.com'),
]

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
LOGS_DIR = os.environ.get('LOGS_DIR', os.path.realpath(os.path.join(BASE_DIR, '../logs')))
if not os.path.isdir(LOGS_DIR):
    os.makedirs(LOGS_DIR)
ML_CLASSIFIERS_DIR = os.environ.get('ML_CLASSIFIERS_DIR', os.path.join(BASE_DIR, '../../ml-classifiers'))
HOST_ML_CLASSIFIERS_DIR = os.environ.get('HOST_ML_CLASSIFIERS_DIR', ML_CLASSIFIERS_DIR)

# File uploads
MEDIA_URL = '/media/'
MEDIA_ROOT = os.environ.get('MEDIA_ROOT', os.path.join(BASE_DIR, 'media'))
HOST_MEDIA_ROOT = os.environ.get('HOST_MEDIA_ROOT', MEDIA_ROOT)
if not os.path.isdir(MEDIA_ROOT):
    os.makedirs(MEDIA_ROOT)

# Local IIIF server
LOCAL_IMAGESERVER_ROOT = os.environ.get('LOCAL_IMAGESERVER_ROOT', os.path.join(BASE_DIR, 'iiif-users'))
LOCAL_IMAGESERVER_ID = int(os.environ.get('LOCAL_IMAGESERVER_ID', 1))
HOST_LOCAL_IMAGESERVER_ROOT = os.environ.get('HOST_LOCAL_IMAGESERVER_ROOT', LOCAL_IMAGESERVER_ROOT)

# Extra initial data for scripts
HOST_INITIAL_DATA_ROOT = os.environ.get('HOST_INITIAL_DATA', os.path.join(BASE_DIR, '../../data'))

# Amazon S3
# S3SOURCE_ variables are from Cantaloupe
AWS_ACCESS_KEY = os.environ.get('S3SOURCE_ACCESS_KEY_ID')
AWS_SECRET_KEY = os.environ.get('S3SOURCE_SECRET_KEY')
AWS_ENDPOINT = os.environ.get('S3SOURCE_ENDPOINT')
AWS_REGION = os.environ.get('AWS_REGION')
AWS_THUMBNAIL_BUCKET = os.environ.get('AWS_THUMBNAIL_BUCKET', 'thumbnails')
AWS_STAGING_BUCKET = os.environ.get('AWS_STAGING_BUCKET', 'staging')
AWS_IIIF_BUCKET = os.environ.get('S3SOURCE_BASICLOOKUPSTRATEGY_BUCKET_NAME', 'iiif')

if 'test' in sys.argv:
    # Overrides for unit tests
    AWS_ACCESS_KEY = 'test'
    AWS_SECRET_KEY = 'test'
    AWS_ENDPOINT = 'http://s3'

# Ponos integration
if os.environ.get('PONOS_TASK'):
    # In a ponos docker task
    PONOS_RECIPE = None
    PONOS_DATA_DIR = '/data'
    HOST_PONOS_DATA_DIR = '/data'
    ML_CLASSIFIERS_DIR = '/arkindex/classifiers'
    MEDIA_ROOT = '/arkindex/staging'
    LOCAL_IMAGESERVER_ROOT = '/arkindex/iiif'

else:
    # As scheduler or dev
    PONOS_DATA_DIR = os.environ.get('PONOS_DATA_DIR', os.path.join(BASE_DIR, '../../ponos/data'))
    HOST_PONOS_DATA_DIR = os.environ.get('HOST_PONOS_DATA_DIR', PONOS_DATA_DIR)
    PONOS_RECIPE = {
        'volumes': {
            # Always expose host configuration
            HOST_ML_CLASSIFIERS_DIR: '/arkindex/classifiers',
            HOST_MEDIA_ROOT: '/arkindex/staging',
            HOST_LOCAL_IMAGESERVER_ROOT: '/arkindex/iiif',
            HOST_INITIAL_DATA_ROOT: '/arkindex/initial',
        },
        'env': {
            # Use docker internal DNS to access databases
            'DB_HOST': 'ark-database',
            'DB_PORT': 5432,
            'DB_NAME': DATABASES['default']['NAME'],
            'DB_USER': DATABASES['default']['USER'],
            'DB_PASSWORD': DATABASES['default']['PASSWORD'],
            'ES_HOST': 'ark-es-01',
            'ARKINDEX_API_INTERNAL_URL': 'http://ark-backend/api/v1/',
            'ARKINDEX_API_TOKEN': os.environ.get('ARKINDEX_API_TOKEN', 'deadbeefTestToken'),
            'LOCAL_IMAGESERVER_ID': LOCAL_IMAGESERVER_ID,
            'S3SOURCE_ACCESS_KEY_ID': AWS_ACCESS_KEY,
            'S3SOURCE_SECRET_KEY': AWS_SECRET_KEY,
            'S3SOURCE_ENDPOINT': AWS_ENDPOINT,
            'S3SOURCE_BASICLOOKUPSTRATEGY_BUCKET_NAME': AWS_IIIF_BUCKET,
            'AWS_THUMBNAIL_BUCKET': AWS_THUMBNAIL_BUCKET,
            'AWS_STAGING_BUCKET': AWS_STAGING_BUCKET,
            'AWS_REGION': AWS_REGION,
        },
        'network': 'arkindex_default',
        'links': {
            'ark-database': None,
            'ark-es-01': None,
            'ark-backend': None,
        }
    }

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.environ.get('SECRET_KEY', 'jf0w^y&ml(caax8f&a1mub)(js9(l5mhbbhosz3gi+m01ex+lo')

# SECURITY WARNING: don't run with debug turned on in production!
ARKINDEX_ENV = os.environ.get('ARKINDEX_ENV', 'dev')
DEBUG = ARKINDEX_ENV == 'dev'

# Run all system checks when starting the server
ALL_CHECKS = os.environ.get('ALL_CHECKS') == 'true' or 'runserver' in sys.argv

ALLOWED_HOSTS = env2list('ALLOWED_HOSTS', default=['localhost', ])

# Required for django-debug-toolbar
INTERNAL_IPS = ['127.0.0.1', '127.0.1.1']

# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    # Tools
    'rest_framework',
    'rest_framework.authtoken',
    'corsheaders',
    'ponos',

    # Our apps
    'arkindex.images',
    'arkindex.documents',
    'arkindex.users',
    'arkindex.dataimport',
]

MIDDLEWARE = [
    'corsheaders.middleware.CorsMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'arkindex.project.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            os.path.join(BASE_DIR, 'templates'),
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'arkindex.project.wsgi.application'


# Password validation
# https://docs.djangoproject.com/en/1.11/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

AUTH_USER_MODEL = 'users.User'

# Internationalization
# https://docs.djangoproject.com/en/1.11/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.11/howto/static-files/
STATIC_URL = '/static/'
STATIC_ROOT = os.environ.get('STATIC_ROOT')
STATICFILES_DIRS = [
]

DEMO_FILES_LIMIT = 50 * 1024 ** 2  # 50 MiB


# API
REST_FRAMEWORK = {
    'DEFAULT_PERMISSION_CLASSES': (
        'arkindex.project.permissions.AllowAny',
    ),
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.SessionAuthentication',
        'rest_framework.authentication.TokenAuthentication',
    ),
    'DEFAULT_PAGINATION_CLASS': 'arkindex.project.pagination.PageNumberPagination',
    'PAGE_SIZE': 20,
}

# Elastic search config
ELASTIC_SEARCH_HOSTS = [
    os.environ.get('ES_HOST', 'localhost'),
]
# The Scroll API is required to go over 10K results
ES_RESULTS_LIMIT = 10000
# ES defaults to three items returned in a nested query if the inner_hits size is not defined
ES_INNER_RESULTS_LIMIT = 6
ES_INDEX_TRANSCRIPTIONS = 'transcriptions'
ES_INDEX_ACTS = 'acts'
ES_INDEX_PAGES = 'pages'

# Silent logger for elasticsearch
logging.getLogger('elasticsearch').setLevel(logging.WARNING)

# InfluxDB API root
INFLUXDB_API_URL = os.environ.get('INFLUXDB_API_URL', 'http://localhost:8086/')

# SSLify proxy
SSLIFY_HOST = os.environ.get('SSLIFY_HOST')

# Use SSL proxy
SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')

# IIIF manifests
IIIF_PRESENTATION_CONTEXT = "http://iiif.io/api/presentation/2/context.json"
IIIF_IMAGE_CONTEXT = "http://iiif.io/api/image/2/context.json"
IIIF_SEARCH_CONTEXT = "http://iiif.io/api/search/0/context.json"
IIIF_IMAGE_SERVICE_PROFILE = "http://iiif.io/api/image/2/level2.json"
IIIF_SEARCH_SERVICE_PROFILE = "http://iiif.io/api/search/0/search"

# Set to True to show transcription annotation lists instead of act annotation lists
IIIF_TRANSCRIPTION_LIST = False

# IIIF manifest download timeout
# See http://docs.python-requests.org/en/master/user/advanced/#timeouts
IIIF_DOWNLOAD_TIMEOUT = (30, 60)

# check_images sample size when checking all servers
CHECK_IMAGES_SAMPLE_SIZE = 20

TRANSCRIPTIONS_IMPORT_QUEUE_SIZE = 25000
TRANSCRIPTIONS_IMPORT_CHUNK_SIZE = 10000

# TEI XSLT file path
TEI_XSLT_PATH = os.path.join(BASE_DIR, 'documents/teitohtml.xsl')

# GitLab OAuth
GITLAB_APP_ID = os.environ.get('GITLAB_APP_ID')
GITLAB_APP_SECRET = os.environ.get('GITLAB_APP_SECRET')

# Cache into memcached
CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.memcached.MemcachedCache',
        'LOCATION': '{}:11211'.format(os.environ.get('CACHE_HOST', 'localhost')),
    }
}

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse',
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'simple'
        },
        'importers': {
            'level': 'INFO',
            'class': 'logging.FileHandler',
            'formatter': 'verbose',
            'filename': os.path.join(LOGS_DIR, 'importers.log'),
        },
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler',
            'include_html': True,
        },
    },
    'loggers': {
        'django': {
            'handlers': ['console', 'mail_admins', ],
            'level': 'INFO',
            'propagate': True,
        },
        'arkindex.documents.importer': {
            'handlers': ['importers'],
            'level': 'DEBUG',
        },
        'arkindex.documents.surface': {
            'handlers': ['importers'],
            'level': 'DEBUG',
        },
    },
    'formatters': {
        'verbose': {
            'format': '[%(levelname)s] %(asctime)s %(module)s: %(message)s'
        },
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
    },
}

# Email
EMAIL_SUBJECT_PREFIX = '[Arkindex {}] '.format(ARKINDEX_ENV)
if os.environ.get('EMAIL_HOST'):
    EMAIL_BACKEND = 'django.core.mail.backends.smtp.EmailBackend'
    EMAIL_HOST = os.environ.get('EMAIL_HOST')
    EMAIL_PORT = os.environ.get('EMAIL_PORT')
    EMAIL_HOST_USER = os.environ.get('EMAIL_HOST_USER')
    DEFAULT_FROM_EMAIL = SERVER_EMAIL = EMAIL_HOST_USER
    EMAIL_HOST_PASSWORD = os.environ.get('EMAIL_HOST_PASSWORD')
    EMAIL_USE_TLS = True
else:
    EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'

# Cookies
CSRF_COOKIE_NAME = 'arkindex.csrf'
CSRF_TRUSTED_ORIGINS = env2list('CSRF_TRUSTED_ORIGINS')
CSRF_COOKIE_DOMAIN = os.environ.get('COOKIE_DOMAIN')

SESSION_COOKIE_NAME = 'arkindex.auth'
SESSION_COOKIE_DOMAIN = os.environ.get('COOKIE_DOMAIN')

CORS_ORIGIN_WHITELIST = env2list('CORS_ORIGIN_WHITELIST', default=[
    'universalviewer.io',
    'localhost:8080',
    '127.0.0.1:8080',
    'localhost:5000',
    '127.0.0.1:5000',
])
CORS_ALLOW_CREDENTIALS = True
CORS_ALLOW_HEADERS = default_headers + (
    'cache-control',  # Allow the frontend to prevent caching some API requests
)
CORS_URLS_REGEX = r'^/(api|ponos)/.*$'

# Docker images used by our ponos workflow
ARKINDEX_APP_IMAGE = os.environ.get('ARKINDEX_APP_IMAGE', 'arkindex-app')
ARKINDEX_ML_IMAGE = os.environ.get('ARKINDEX_ML_IMAGE', 'arkindex-worker-ml')

# ML worker
ML_DEFAULT_CLASSIFIER = 'tobacco'
ML_DEFAULT_RECOGNIZER = 'tesseract'

# User groups with special permissions
DEMO_GROUP_ID = 1
INTERNAL_GROUP_ID = 2

# Optional unit tests runner with code coverage
try:
    import django_nose # noqa
    INSTALLED_APPS.append('django_nose')
    TEST_RUNNER = 'django_nose.NoseTestSuiteRunner'
    NOSE_ARGS = [
        '--with-coverage',
        '--cover-package=arkindex',
    ]
except ImportError:
    pass

# Docker container name resolution
INTERNAL_HOSTS = ['127.0.0.1', 'localhost', 'backend', 'ark-backend']
ALLOWED_HOSTS += INTERNAL_HOSTS

# Local settings
try:
    from .local_settings import * # noqa
except ImportError:
    pass

try:
    import django_extensions  # noqa
    INSTALLED_APPS.append('django_extensions')
except ImportError:
    pass

try:
    import debug_toolbar # noqa
    MIDDLEWARE.insert(0, 'debug_toolbar.middleware.DebugToolbarMiddleware')
    INSTALLED_APPS.append('debug_toolbar')
except ImportError:
    pass
