from unittest.mock import patch, call
from django.test import TestCase, override_settings
from django.conf import settings
from django.urls import path
from django.core.checks import Error, Warning
from subprocess import CalledProcessError
import tempfile
import shutil
import subprocess


@override_settings(ALL_CHECKS=True)
class ChecksTestCase(TestCase):

    def test_api_urls_check(self):
        '''
        Check should always be OK
        '''
        from arkindex.project.checks import api_urls_check
        from arkindex.project.api_v1 import api

        self.assertListEqual(api_urls_check(), [])

        # Add an invalid url
        bad_url = path('bad', lambda: None, name='bad-url')
        api.append(bad_url)

        # Should trigger the check
        self.assertListEqual(
            api_urls_check(),
            [
                Error(
                    'This api url misses a mandatory ending /',
                    hint=str(bad_url),
                    id='arkindex.E001',
                )
            ]
        )

    @patch('arkindex_common.ml_tool.MLTool.get')
    @patch('arkindex.project.checks.os.path.isdir')
    def test_ml_default_tools_check(self, isdir_mock, ml_get_mock):
        """
        Test the default ML tools existence checks
        """
        from arkindex.project.checks import ml_default_tools_check

        isdir_mock.return_value = True
        self.assertListEqual(ml_default_tools_check(), [])

        isdir_mock.return_value = False
        ml_get_mock.side_effect = ValueError

        with self.settings(
                ML_CLASSIFIERS_DIR='oops',
                ML_DEFAULT_CLASSIFIER='fail1',
                ML_DEFAULT_RECOGNIZER='fail2'):
            self.assertListEqual(
                ml_default_tools_check(),
                [
                    Error(
                        'ML tools directory does not exist',
                        hint='settings.ML_CLASSIFIERS_DIR = "oops"',
                        id='arkindex.E008',
                    ),
                    Error(
                        'Default Classifier does not exist',
                        hint='settings.ML_DEFAULT_CLASSIFIER = "fail1"',
                        id='arkindex.E003',
                    ),
                    Error(
                        'Default Recognizer does not exist',
                        hint='settings.ML_DEFAULT_RECOGNIZER = "fail2"',
                        id='arkindex.E003',
                    ),
                ]
            )

    def test_local_imageserver_check(self):
        """
        Test the local imageserver existence check
        """
        from arkindex.project.checks import local_imageserver_check
        from arkindex.images.models import ImageServer
        self.assertEqual(ImageServer.objects.count(), 0)

        with self.settings(LOCAL_IMAGESERVER_ID=42, LOCAL_IMAGESERVER_ROOT='/nowhere'):
            self.assertListEqual(local_imageserver_check(), [
                Error(
                    'Local ImageServer with ID 42 does not exist',
                    hint='settings.LOCAL_IMAGESERVER_ID = 42',
                    id='arkindex.E004',
                ),
                Error(
                    'Local ImageServer root folder does not exist',
                    hint='settings.LOCAL_IMAGESERVER_ROOT = "/nowhere"',
                    id='arkindex.E005',
                ),
            ])

        srv = ImageServer.objects.create(display_name="Test Local Server", url="http://nowhere")
        with self.settings(LOCAL_IMAGESERVER_ID=srv.id, LOCAL_IMAGESERVER_ROOT='/nowhere'):
            self.assertListEqual(local_imageserver_check(), [
                Error(
                    'Local ImageServer root folder does not exist',
                    hint='settings.LOCAL_IMAGESERVER_ROOT = "/nowhere"',
                    id='arkindex.E005',
                ),
            ])

        folder = tempfile.mkdtemp()
        with self.settings(LOCAL_IMAGESERVER_ID=srv.id, LOCAL_IMAGESERVER_ROOT=folder):
            self.assertListEqual(local_imageserver_check(), [])

        srv.delete()
        shutil.rmtree(folder)

    @patch('arkindex.project.checks.subprocess.run')
    def test_docker_images_check(self, run_mock):
        from arkindex.project.checks import docker_images_check

        run_mock.side_effect = [CalledProcessError(1, ''), None]
        with self.settings(ARKINDEX_APP_IMAGE='nope', ARKINDEX_ML_IMAGE='me-neither'):
            self.assertListEqual(docker_images_check(), [
                Error(
                    'Docker image with tag "nope" was not found.',
                    hint='settings.ARKINDEX_APP_IMAGE = "nope"',
                    id='arkindex.E006',
                )
            ])

        self.assertEqual(run_mock.call_count, 2)
        self.assertEqual(run_mock.call_args_list, [
            call(
                ['docker', 'image', 'inspect', 'nope'],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                check=True,
            ),
            call(
                ['docker', 'image', 'inspect', 'me-neither'],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                check=True,
            ),
        ])

        run_mock.reset_mock()
        run_mock.side_effect = [None, CalledProcessError(1, '')]
        with self.settings(ARKINDEX_APP_IMAGE='nope', ARKINDEX_ML_IMAGE='me-neither'):
            self.assertListEqual(docker_images_check(), [
                Error(
                    'Docker image with tag "me-neither" was not found.',
                    hint='settings.ARKINDEX_ML_IMAGE = "me-neither"',
                    id='arkindex.E006',
                )
            ])

        self.assertEqual(run_mock.call_count, 2)
        self.assertEqual(run_mock.call_args_list, [
            call(
                ['docker', 'image', 'inspect', 'nope'],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                check=True,
            ),
            call(
                ['docker', 'image', 'inspect', 'me-neither'],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                check=True,
            ),
        ])

    @patch('arkindex.project.checks.subprocess.run')
    def test_docker_images_check_missing_client(self, run_mock):
        """
        Test the Docker images check does not show errors if the Docker client is missing
        """
        from arkindex.project.checks import docker_images_check

        run_mock.side_effect = FileNotFoundError
        with self.settings(ARKINDEX_APP_IMAGE='nope', ARKINDEX_ML_IMAGE='me-neither'):
            self.assertListEqual(docker_images_check(), [])

        self.assertEqual(run_mock.call_count, 2)
        self.assertEqual(run_mock.call_args_list, [
            call(
                ['docker', 'image', 'inspect', 'nope'],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                check=True,
            ),
            call(
                ['docker', 'image', 'inspect', 'me-neither'],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                check=True,
            ),
        ])

    @patch('arkindex.project.checks.parse_recipe')
    def test_ponos_recipe_check(self, parse_mock):
        from arkindex.project.checks import ponos_recipe_check
        self.assertListEqual(ponos_recipe_check(), [])

        parse_mock.side_effect = AssertionError('oh no')
        self.assertListEqual(ponos_recipe_check(), [Error(
            'Base Ponos recipe validation failed: oh no',
            hint='settings.PONOS_RECIPE',
            id='arkindex.E007',
        )])

    @override_settings()
    def test_ponos_host_check(self):
        from arkindex.project.checks import ponos_host_check
        self.assertListEqual(ponos_host_check(), [])
        del settings.HOST_ML_CLASSIFIERS_DIR
        del settings.HOST_MEDIA_ROOT
        del settings.HOST_LOCAL_IMAGESERVER_ROOT
        del settings.HOST_INITIAL_DATA_ROOT
        self.assertListEqual(ponos_host_check(), [
            Error(
                'HOST_ML_CLASSIFIERS_DIR must be set to a path',
                hint='settings.HOST_ML_CLASSIFIERS_DIR',
                id='arkindex.E009',
            ),
            Error(
                'HOST_MEDIA_ROOT must be set to a path',
                hint='settings.HOST_MEDIA_ROOT',
                id='arkindex.E009',
            ),
            Error(
                'HOST_LOCAL_IMAGESERVER_ROOT must be set to a path',
                hint='settings.HOST_LOCAL_IMAGESERVER_ROOT',
                id='arkindex.E009',
            ),
            Error(
                'HOST_INITIAL_DATA_ROOT must be set to a path',
                hint='settings.HOST_INITIAL_DATA_ROOT',
                id='arkindex.E009',
            ),
        ])

    @patch('arkindex.project.checks.os.path.isdir')
    def test_ponos_data_dir_check(self, isdir_mock):
        from arkindex.project.checks import ponos_data_dir_check

        isdir_mock.return_value = True
        self.assertListEqual(ponos_data_dir_check(), [])

        isdir_mock.return_value = False
        with self.settings(PONOS_DATA_DIR="/oh/no", HOST_PONOS_DATA_DIR="/no/oh"):
            self.assertListEqual(ponos_data_dir_check(), [
                Error(
                    'Ponos working directory does not exist',
                    hint='settings.PONOS_DATA_DIR = "/oh/no"',
                    id='arkindex.E010',
                ),
                Error(
                    'Ponos working directory does not exist on host',
                    hint='settings.HOST_PONOS_DATA_DIR = "/no/oh"',
                    id='arkindex.E010',
                ),
            ])

    @override_settings()
    def test_internal_group_check(self):
        from arkindex.project.checks import internal_group_check
        from django.contrib.auth.models import Group

        del settings.INTERNAL_GROUP_ID
        self.assertListEqual(internal_group_check(), [Warning(
            'Internal group ID is not set; internal API requests will fail.',
            hint='settings.INTERNAL_GROUP_ID',
            id='arkindex.W001',
        )])

        settings.INTERNAL_GROUP_ID = 1234
        self.assertListEqual(internal_group_check(), [Warning(
            'Internal user group does not exist; internal API requests will fail.',
            hint='settings.INTERNAL_GROUP_ID = "1234"',
            id='arkindex.W002',
        )])

        Group.objects.create(id=1234)
        self.assertListEqual(internal_group_check(), [])

    @override_settings()
    def test_demo_group_check(self):
        from arkindex.project.checks import demo_group_check
        from django.contrib.auth.models import Group

        del settings.DEMO_GROUP_ID
        self.assertListEqual(demo_group_check(), [Warning(
            'Demo user group ID is not set; demo registration will fail.',
            hint='settings.DEMO_GROUP_ID',
            id='arkindex.W003',
        )])

        settings.DEMO_GROUP_ID = 1234
        self.assertListEqual(demo_group_check(), [Warning(
            'Demo user group does not exist; demo registration will fail.',
            hint='settings.DEMO_GROUP_ID = "1234"',
            id='arkindex.W004',
        )])

        Group.objects.create(id=1234)
        self.assertListEqual(demo_group_check(), [])

    @override_settings()
    def test_gitlab_oauth_check(self):
        from arkindex.project.checks import gitlab_oauth_check

        del settings.GITLAB_APP_ID
        del settings.GITLAB_APP_SECRET

        self.assertListEqual(gitlab_oauth_check(), [
            Warning(
                'GitLab app ID is not set; Git imports will fail.',
                hint='settings.GITLAB_APP_ID = None',
                id='arkindex.W003',
            ),
            Warning(
                'GitLab app secret is not set; Git imports will fail.',
                hint='settings.GITLAB_APP_SECRET = None',
                id='arkindex.W004',
            ),
        ])

        settings.GITLAB_APP_ID = '1234'
        settings.GITLAB_APP_SECRET = 's3kr3t'

        self.assertListEqual(gitlab_oauth_check(), [])

    @override_settings()
    def test_s3_check(self):
        from arkindex.project.checks import s3_check

        del settings.AWS_ACCESS_KEY
        del settings.AWS_SECRET_KEY
        del settings.AWS_ENDPOINT
        del settings.AWS_THUMBNAIL_BUCKET
        del settings.AWS_STAGING_BUCKET
        del settings.AWS_IIIF_BUCKET
        self.maxDiff = None
        self.assertCountEqual(s3_check(), [
            Error(
                'AWS access key ID is missing; all S3-related features will fail.',
                hint='settings.AWS_ACCESS_KEY = None',
                id='arkindex.E011',
            ),
            Error(
                'AWS secret key is missing; all S3-related features will fail.',
                hint='settings.AWS_SECRET_KEY = None',
                id='arkindex.E011',
            ),
            Error(
                'AWS endpoint is missing; all S3-related features will fail.',
                hint='settings.AWS_ENDPOINT = None',
                id='arkindex.E011',
            ),
            Error(
                'S3 thumbnails bucket name is missing; all S3-related features will fail.',
                hint='settings.AWS_THUMBNAIL_BUCKET = None',
                id='arkindex.E011',
            ),
            Error(
                'S3 staging bucket name is missing; all S3-related features will fail.',
                hint='settings.AWS_STAGING_BUCKET = None',
                id='arkindex.E011',
            ),
            Error(
                'S3 IIIF bucket name is missing; all S3-related features will fail.',
                hint='settings.AWS_IIIF_BUCKET = None',
                id='arkindex.E011',
            ),
        ])

        settings.AWS_ACCESS_KEY = 'key'
        settings.AWS_SECRET_KEY = 's3kr3t'
        settings.AWS_ENDPOINT = 'http://somewhere'
        settings.AWS_THUMBNAIL_BUCKET = 'Thumbs.db'
        settings.AWS_STAGING_BUCKET = 'buckette'
        settings.AWS_IIIF_BUCKET = 'bucks'
        self.assertListEqual(s3_check(), [])
