from rest_framework import serializers
from django.utils.http import urlsafe_base64_decode
from django.contrib.auth.tokens import default_token_generator
from arkindex.users.models import OAuthCredentials, OAuthStatus, User
from arkindex.project.serializer_fields import EnumField


class OAuthCredentialsSerializer(serializers.ModelSerializer):

    provider_display_name = serializers.CharField(source='provider_class.display_name')
    status = EnumField(OAuthStatus)

    class Meta:
        model = OAuthCredentials
        fields = (
            'id',
            'status',
            'provider_name',
            'provider_display_name',
            'provider_url',
            'account_name',
        )


class OAuthProviderClassSerializer(serializers.Serializer):

    name = serializers.CharField(source='slug')
    display_name = serializers.CharField()
    default_url = serializers.URLField(source='url')


class OAuthRetrySerializer(serializers.Serializer):
    """
    A serializer used by the OAuthRetry endpoint to return an authorization URL.
    Required to get the OpenAPI schema generation to work.
    """

    url = serializers.URLField()


class UserSerializer(serializers.ModelSerializer):

    class Meta:
        model = User
        fields = (
            'id',
            'email',
            'password',
            'is_admin',
            'auth_token',
        )
        extra_kwargs = {
            'id': {'read_only': True},
            'email': {'read_only': True},
            'password': {'write_only': True},
            'is_admin': {'read_only': True},
            'auth_token': {'read_only': True},
        }

    def update(self, instance, validated_data):
        if 'password' in validated_data:
            instance.set_password(validated_data.pop('password'))
        return super().update(instance, validated_data)


class NewDemoUserSerializer(serializers.ModelSerializer):

    class Meta:
        model = User
        fields = (
            'email',
        )

    def create(self, validated_data):
        return User.objects.create_demo_user(validated_data['email'])


class EmailLoginSerializer(serializers.Serializer):

    email = serializers.EmailField()
    password = serializers.CharField(write_only=True)


class PasswordResetSerializer(serializers.Serializer):
    """
    A serializer that only takes an e-mail as the input.
    This could use a ModelSerializer on User; however, just as the Django
    default password reset view does, there should not be any errors when
    asking for a password reset on a wrong e-mail, because that would give
    account existence information to attackers.
    """
    email = serializers.EmailField()


class PasswordResetConfirmSerializer(serializers.Serializer):
    """
    Second step of a password reset: when the user clicked the link in the email.
    The frontend will send the base64-encoded user ID, reset token and new password.
    """
    uidb64 = serializers.CharField(write_only=True)
    token = serializers.CharField(write_only=True)
    password = serializers.CharField(write_only=True)

    def validate(self, data):
        # Never fail, even with an invalid UUID or non-existent user
        try:
            pk = urlsafe_base64_decode(data['uidb64']).decode()
            user = User.objects.get(pk=pk)
        except (TypeError, ValueError, User.DoesNotExist):
            user = None

        valid_token = user is not None and default_token_generator.check_token(user, data['token'])

        return {'user': user, 'valid_token': valid_token, 'password': data['password']}

    def save(self):
        user = self.validated_data['user']
        if not user or not self.validated_data['valid_token']:
            return
        user.set_password(self.validated_data['password'])
        user.save()
