from rest_framework import status
from django.urls import reverse
from django.core import mail
from django.contrib import auth
from django.contrib.auth.tokens import default_token_generator
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import User
import urllib.parse


class TestRegistration(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = User.objects.create_user('email@address.com', 'P4$5w0Rd')

    def test_retrieve_requires_login(self):
        response = self.client.get(reverse('api:user-retrieve'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_requires_login(self):
        response = self.client.put(
            reverse('api:user-retrieve'),
            data={'password': 'N€wP4$5w0Rd'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logout_requires_login(self):
        response = self.client.delete(reverse('api:user-retrieve'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retrieve(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:user-retrieve'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['email'], 'email@address.com')
        self.assertEqual(data['is_admin'], False)
        self.assertNotIn('password', data)

    def test_update(self):
        self.client.force_login(self.user)
        response = self.client.put(
            reverse('api:user-retrieve'),
            data={'password': 'N€wP4$5w0Rd'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.user.refresh_from_db()
        self.assertTrue(self.user.check_password('N€wP4$5w0Rd'))

    def test_logout(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:user-retrieve'))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(auth.get_user(self.client).is_authenticated)

    def test_create(self):
        response = self.client.post(
            reverse('api:user-new'),
            data={'email': 'newuser@example.com'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertTrue(auth.get_user(self.client).is_authenticated)
        self.assertEqual(auth.get_user(self.client).email, 'newuser@example.com')
        self.assertFalse(User.objects.get_by_natural_key('newuser@example.com').verified_email)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].to, ['newuser@example.com'])

    def test_login_failure(self):
        response = self.client.post(
            reverse('api:user-login'),
            data={'email': 'email@address.com', 'password': 'wrongpassword'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertFalse(auth.get_user(self.client).is_authenticated)

    def test_login(self):
        response = self.client.post(
            reverse('api:user-login'),
            data={'email': 'email@address.com', 'password': 'P4$5w0Rd'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertTrue(auth.get_user(self.client).is_authenticated)
        self.assertEqual(auth.get_user(self.client).email, 'email@address.com')

    def test_email_verification(self):
        newuser = User.objects.create_user('newuser@example.com')
        self.assertFalse(newuser.verified_email)
        self.assertFalse(newuser.has_usable_password())

        response = self.client.get('{}?{}'.format(
            reverse('api:user-token'),
            urllib.parse.urlencode({
                'email': 'newuser@example.com',
                'token': default_token_generator.make_token(newuser),
            })),
        )
        self.assertEqual(response.status_code, status.HTTP_302_FOUND)

        newuser.refresh_from_db()
        self.assertTrue(newuser.verified_email)
