#!/usr/bin/env python3
import argparse
import apistar
import os.path
import pkg_resources
import sys
import yaml

BASE_DIR = os.path.dirname(os.path.abspath(__file__))


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'patch',
        help='File describing patches to make on the schema',
        type=argparse.FileType('r'),
    )
    parser.add_argument(
        'original',
        help='Generated OpenAPI schema',
        type=argparse.FileType('r'),
        nargs='?',
        default=sys.stdin,
    )
    parser.add_argument(
        'output',
        help='Destination file',
        type=argparse.FileType('w'),
        nargs='?',
        default=sys.stdout,
    )
    args = vars(parser.parse_args())

    # Load YAML files
    args['original'] = yaml.load(args['original'])
    args['patch'] = yaml.load(args['patch'])
    return args


def fix_nested_properties(schema):
    """
    Walk through the schema recursively, looking for Schema Objects with type=object,
    to change {"properties": {"properties": {}, "required": []}
    into {"properties": {}, "required": []}
    See https://github.com/encode/django-rest-framework/pull/6568
    """
    if schema.get('type') == 'object' \
            and 'properties' in schema \
            and 'properties' in schema['properties'] \
            and 'required' in schema['properties']:
        schema = fix_nested_properties(schema['properties'])
        schema['type'] = 'object'
        return schema

    return {
        k: (fix_nested_properties(v) if isinstance(v, dict) else v)
        for k, v in schema.items()
    }


def update_schema(schema, patches):
    """
    Perform updates on an OpenAPI schema using another YAML file describing patches.
    The update is not recursive; any root key will overwrite the schema's original data.
    For paths, each operation has its keys updated; the whole operation is not overwritten.
    """
    # Update each method separately
    paths = patches.pop('paths', {})
    for path, methods in paths.items():
        if path not in schema['paths']:
            print('Creating path {}'.format(path), file=sys.stderr)
            schema['paths'][path] = {}

        for method, operation in methods.items():
            print(
                '{} method {} on path {}'.format(
                    'Updating' if method in schema['paths'][path] else 'Creating',
                    method,
                    path,
                ),
                file=sys.stderr,
            )
            schema['paths'][path].setdefault(method, {})
            schema['paths'][path][method].update(operation)

    # Update other root keys
    for k, v in patches.items():
        print('Patching {}'.format(k), file=sys.stderr)
        schema[k] = v

    # Set the API version to the Arkindex package version
    try:
        # Try with the VERSION file
        with open(os.path.join(BASE_DIR, '..', 'VERSION')) as f:
            ark_version = f.read().strip()
    except FileNotFoundError:
        # Fall back to the pip package version
        ark_version = pkg_resources.get_distribution('arkindex').version

    schema.setdefault('info', {})
    schema['info']['version'] = ark_version

    # Add x-name property on request body objects to let apistar register the parameters
    for path, methods in schema['paths'].items():
        for method, operation in methods.items():
            if 'requestBody' not in operation:
                continue
            if not operation['requestBody']['content']['application/json']['schema']:
                # Ignore empty schemas
                continue

            # Make sure we do not duplicate an existing parameter name
            assert not any(param['name'] == 'body' for param in operation['parameters']), \
                'Operation already has a body parameter'

            print('Adding x-name to {} on {}'.format(method, path), file=sys.stderr)
            operation['requestBody']['x-name'] = 'body'

    return fix_nested_properties(schema)


def print_endpoints(schema):
    """
    Output a list of all endpoints with their operation IDs and security settings to stderr
    """
    for path, methods in schema['paths'].items():
        print(path, file=sys.stderr)
        for method, operation in methods.items():
            security = operation.get('security')
            auth = 'Custom'
            if security is None:
                auth = 'Default'
            elif security == []:
                auth = 'No'

            print('  {}: {} - {} authentication'.format(
                method,
                operation.get('operationId', 'No operation ID!'),
                auth,
            ), file=sys.stderr)


def main():
    args = get_args()
    schema = update_schema(args['original'], args['patch'])
    print_endpoints(schema)
    apistar.validate(schema)
    yaml.dump(schema, args['output'], default_flow_style=False)


if __name__ == '__main__':
    main()
